"""
section.py: Section class
"""

from ..pandocastobject.pandocast.element import Element
from .textblock import TextBlock


class Section(list):
    """ """

    def __init__(self, iterable=[], level=0, opts={}):
        super().__init__(iterable)
        self.level = level
        self.__opts = opts  # not yet copy() / copy.deepcopy()
        TEXT_BLOCK_TYPES = (
            self.__opts.get("pandocast", {}).get("types", {}).get("textblock")
        )

        block: Element = None
        i = 0
        while i < len(self):
            block = self[i]
            block_type = block.get_type()

            if (block_type == "Header") and (not (self.level > 0 and i == 0)):

                lv = block.get_prop("Level")
                if self.level > 0:
                    if (lv < self.level) or ((i > 0) and (lv == self.level)):

                        raise RuntimeError()

                sublevel = lv
                subend = i + 1
                while subend < len(self):
                    subblock: Element = self[subend]
                    if (subblock.get_type() == "Header") and (
                        lv := subblock.get_prop("Level")
                    ) <= sublevel:
                        break
                    subend += 1

                section = Section(self[i:subend], sublevel, self.__opts)
                self[i:subend] = [section]

            elif block_type in TEXT_BLOCK_TYPES:
                self[i] = TextBlock(block)

            # # Setup Gdoc data class here
            # else:
            #     self[i] = BlockClassTypes(self[i])

            i += 1

    def __getitem__(self, index):
        result = super().__getitem__(index)
        # if type(index) is slice:
        #     result = Section(result, self.level, self.__opts)

        return result



# SysML plugin
#
# 1. Requirement diagram 相当の情報を取り扱う
#

import logging

from ... import debug, gdom

_LOGGER = logging.getLogger(__name__)
_DEBUG = debug.Debug(_LOGGER)


class HierarchicalDict(gdom.Object):
    def __init__(self, table, tag=None, parent=None) -> None:
        _DEBUG.print("class HierarchicalDict(gdom.Object) {")
        _DEBUG.indent()

        super().__init__(table, tag)

        self.table = table
        self.content = [{}, {}]  # Set of Properties dict and Children dict.

        # Step.1: header から、列：要素名の対応リストを作成する
        #         1列名はとりあえず衝突回避のために "" 固定（Keyに採用されるので出力されない）
        headerKey = []
        cell = table.getCell(1, 1)
        while cell is not None:
            line = cell.getFirstLine()
            headerKey.append(line)
            cell = cell.next()
        _DEBUG.print("headerKey = [" + ", ".join(headerKey) + "]")
        headerKey[0] = ""

        # ASTのヘッダ行数を確認し、データ先頭行を取得
        dataStartRow = table.numHeaderRows
        if dataStartRow == 0:
            dataStartRow = table.getCell(1, 1).getProp("RowSpan")
        dataStartRow += 1

        dataEndRow = table.numTableRows - table.numFooterRows
        _DEBUG.print("dataStartRow = " + str(dataStartRow))
        _DEBUG.print("dataEndRow = " + str(dataEndRow))

        next = dataStartRow

        while next <= dataEndRow:
            data = [{}, {}]
            next = self._parser(next, dataEndRow, headerKey, 0, data)
            self.content[1][data[0]["__key__"]] = data

        _DEBUG.undent()
        _DEBUG.print("}")

    # 暫定：このクラスは持つべきでないメソッド。あとで削除する。
    def getItems(self, symboltable):
        return []

    def getProp(self, ids, key, element=None):
        content = self.content if element is None else element

        if ids is not None:
            for id in ids:
                content = content[1][id]

        return content[0].get(key)

    def setProp(self, ids, key, val, element=None):
        content = self.content if element is None else element

        if ids is not None:
            for id in ids:
                content = content[1][id]

        content[0][key] = val

    def getChildren(self, element=None):
        content = self.content if element is None else element
        return content[1]

    def getElement(self, ids=None, element=None):
        content = self.content if element is None else element

        if ids is not None:
            for id in ids:
                content = content[1][id]

        return content

    def _parser(self, start, end, headerKey, top, data):
        isFirstRow = True
        data[0]["__key__"] = ""

        next = start
        while next <= end:

            # Skip empty row
            row = self._getRowData(next)
            if row is None:
                next += 1
                continue

            # return if parent top data exists.
            for c in range(0, top - 1):
                if row[c] != "":
                    return next

            # switch(row type)
            # case keyword-line
            if row[top] == "":
                key = None
                val = None
                obj = None

                col = top + 1
                while col < len(row):
                    if row[col] != "":
                        key = row[col]
                        break
                    col += 1

                if key == "@":
                    obj = [{}, {}]
                    next = self._parseSubItem(self, next, end, headerKey, col, obj)
                    key = obj[0]["__key__"]

                elif key is not None:
                    for c in range(col + 1, len(row)):
                        if val is None:
                            val = [row[c]]
                        else:
                            val.append(row[c])

                if key is not None:
                    if val is not None:
                        data[0][key] = val
                        isFirstRow = False

                    elif obj is not None:
                        data[1][key] = obj
                        isFirstRow = False

                    else:
                        return "ERROR: Syntax error. unexpected end of row."

            # case subid行
            elif row[top] == "@":
                cell = row[top + 1]
                if cell == "":
                    return "ERROR: Syntax error. Key string is missing after '@'."

                data[1][cell] = [{}, {}]

                next = self._parseSubItem(next, end, headerKey, top, data[1][cell])
                continue

            # case id
            elif isFirstRow:
                for c in range(top + 1, len(row)):
                    data[0][headerKey[c]] = row[c]

                data[0]["__key__"] = row[top]
                isFirstRow = False

            # case next id
            else:
                break

            next += 1

        return next

    def _parseSubItem(self, start, end, headerKey, top, data):
        isFirstRow = True

        next = start
        while next <= end:

            # Skip empty row
            row = self._getRowData(next)
            if row is None:
                next += 1
                continue

            # return if parent top data exists.
            for c in range(0, top - 1):
                if row[c] != "":
                    return next

            # switch(row type)
            # case keyword-line
            if row[top] == "":
                key = None
                val = None
                obj = None

                col = top + 1
                while col < len(row):
                    if row[col] != "":
                        key = row[col]
                        break
                    col += 1

                if key == "@":
                    obj = [{}, {}]
                    next = self._parseSubItem(next, end, headerKey, col, obj)
                    key = obj[0]["__key__"]

                elif key is not None:
                    for c in range(col + 1, len(row)):
                        if val is None:
                            val = [row[c]]
                        else:
                            val.append(row[c])

                if key is not None:
                    if val is not None:
                        data[0][key] = val
                        isFirstRow = False

                    elif obj is not None:
                        data[1][key] = obj
                        isFirstRow = False

                    else:
                        return "ERROR: Syntax error. unexpected end of row."

            # case '@'tagged row
            elif (row[top] == "@") and isFirstRow:
                cell = row[top + 1]
                if cell == "":
                    return "ERROR: Syntax error. Key string is missing after '@'."

                for c in range(top + 2, len(row)):
                    data[0][headerKey[c]] = row[c]

                data[0]["__key__"] = row[top + 1]
                isFirstRow = False

            # case id or next'@'
            else:
                break

            next += 1

        return next

    def _isId(self, cell):
        isid = (not cell.isEmpty()) and (not (cell.getFirstLine() in ["@", "#"]))
        return isid

    def _getRowData(self, row):
        rowdata = []
        isComment = False
        # row = self.table.getRow(row)

        for c in range(1, self.table.numTableColumns + 1):
            # line = row.getCell(c).getFirstLine()
            line = self.table.getCell(row, c).getFirstLine()
            if line == "#":
                isComment = True
            line = "" if isComment else line
            rowdata.append(line)

        isEmpty = True
        for c in range(0, self.table.numTableColumns):
            if rowdata[c] != "":
                isEmpty = False

        rowdata = None if isEmpty else rowdata

        return rowdata


exports = {"Table": {"HierarchicalDict": {"constructor": HierarchicalDict}}}



r"""
GdSymbol class
"""
import re

from gdoc.lib.gdoccompiler.gdexception import GdocSyntaxError


class GdSymbol:
    """
    ;
    """

    IS_NAME_STR = "@"
    # "@" is a character not allowed in id.
    # Set it at the beginning of a string to indicate that
    # the string is not an id, i.e., a name.

    def __init__(self, symbol):
        """ """
        self.__symbols, self.__tags = GdSymbol._split_symbol(symbol)
        self.__symbol_str = symbol  # str is immutable and needless to copy.
        # TODO: PandocStr needs to copy.

    def is_id(self):
        """ """
        return not self.__symbols[-1].startswith(GdSymbol.IS_NAME_STR)

    def get_symbols(self):
        """ """
        return self.__symbols[:]

    def get_symbol_str(self):
        """ """
        symbol_str = ""

        for s in self.__symbols:
            if s.startswith(GdSymbol.IS_NAME_STR):
                symbol_str += "[" + s[1:] + "]"
            elif symbol_str != "":
                symbol_str += "." + s
            else:
                symbol_str = s

        return symbol_str

    def get_tags(self):
        """ """
        return self.__tags[:]

    #
    # Class Methods
    #
    @classmethod
    def is_valid_symbol(cls, symbol):
        """ """
        result = True

        try:
            cls._split_symbol(symbol)
        except:
            result = False

        return result

    @classmethod
    def is_valid_id(cls, symbol):
        """ """
        result = GdSymbol._is_gdoc_identifier(symbol)

        if result is not True:
            result = False

        return result

    #
    # Class Private Variables and Methods
    #
    _IDENTIFIER = re.compile(r"\.(\S*?)(?=(\.|\[|\(|$))")
    _IDENTIFIER_INDEX = 1

    _TAG_STRING = re.compile(r"\((.*)\)$")
    _TAG_STRING_INDEX = 1
    _TAG_DELIMITER = re.compile(r"\s*,\s*|\s+")
    _TAG_INVALID_SYNTAX = re.compile(r"\(\s*,|,\s*[,)]")

    @classmethod
    def _split_symbol(cls, symbolstr: str):
        """split symbol string to ids or names and tags.
        @param symbol : str | PandocStr
        @return (list(str), list(str))

        T.B.D. : Error handling
        """
        symbol = symbolstr.strip()
        column = symbolstr.index(symbol) + 1  # Column is starting from 1.

        if not symbol.startswith(("[", "(")):
            target = "." + symbol
            column -= 1
        else:
            target = symbol

        symbols = []
        tags = []
        while len(target) > 0:
            if target.startswith("["):
                try:
                    name, length = cls._get_namestr(target)
                except GdocSyntaxError as err:
                    raise GdocSyntaxError(
                        err.msg, (None, None, column + err.offset, str(symbolstr))
                    ) from err

                if name.startswith('"'):
                    name = name[1:-1]
                else:
                    if (i := cls._is_gdoc_identifier(name)) is not True:
                        raise GdocSyntaxError(
                            "invalid character in non-quoted name",
                            (None, None, column + i + 1, str(symbolstr))
                            # column + i + len('[')
                        )

                symbols.append(GdSymbol.IS_NAME_STR + name)

            elif s := cls._IDENTIFIER.match(target):
                identifier = s.group(cls._IDENTIFIER_INDEX)
                length = len(s.group(0))

                if identifier == "":
                    column = column + length
                    if column > len(symbolstr):
                        # in case _IDENTIFIER matched with $(EOL)
                        column -= 1

                    raise GdocSyntaxError(
                        "invalid syntax", (None, None, column, str(symbolstr))
                    )
                elif (i := cls._is_gdoc_identifier(identifier)) is not True:
                    raise GdocSyntaxError(
                        "invalid character in identifier",
                        (None, None, column + i + 1, str(symbolstr))
                        # column + i + len('.')
                    )

                symbols.append(identifier)

            elif s := cls._TAG_STRING.match(target):
                tags = s.group(cls._TAG_STRING_INDEX).strip()
                length = len(s.group(0))

                if len(tags) > 0:
                    tags = cls._TAG_DELIMITER.split(tags)
                else:
                    tags = []

                for tag in tags:
                    if tag != "":
                        hash_added = 1 if tag.startswith("#") else 0
                        if (i := cls._is_gdoc_identifier(tag[hash_added:])) is not True:
                            i += target.index(tag) + hash_added
                            raise GdocSyntaxError(
                                "invalid character in tag",
                                (None, None, column + i, str(symbolstr)),
                            )
                    else:  # tag == ""
                        m = cls._TAG_INVALID_SYNTAX.search(target)
                        raise GdocSyntaxError(
                            "invalid syntax",
                            (None, None, column + m.end() - 1, str(symbolstr)),
                        )

            else:
                raise GdocSyntaxError(
                    "invalid syntax", (None, None, column, str(symbolstr))
                )

            column += length
            target = target[length:]

        return symbols, tags

    @classmethod
    def _is_gdoc_identifier(cls, identifier):
        result = False

        if ("a" + identifier).isidentifier():
            result = True

        else:
            for c in identifier:  # pragma: no branch: This line never complete.
                if not ("a" + c).isidentifier():
                    break

            result = identifier.index(c)

        return result

    @classmethod
    def _get_namestr(cls, target):
        name = ""
        index = 1  # skip the top '['
        length = len(target)

        #
        # Skip leading white spaces
        #
        while index < length and target[index] == " ":
            index += 1

        if index == length:
            raise GdocSyntaxError(
                "EOS while scanning name string", (None, None, index - 1, target)
            )

        #
        # Quoted string
        #
        if target[index] == '"':
            name += '"'
            index += 1

            while index < length:
                c = target[index]

                if c == '"':  # Wait closing '"'
                    name += c
                    index += 1
                    break

                if c == "\\":  # Escape sequence
                    index += 1
                    if index == length:
                        break

                    c = target[index]
                    if c == '"':
                        name += '"'
                    elif c == "\\":
                        name += "\\"
                    else:
                        name += "\\" + c

                else:
                    name += c

                index += 1

            if index == length:  # String not close
                raise GdocSyntaxError(
                    "EOS while scanning string literal", (None, None, index - 1, target)
                )

        #
        # invalid syntax
        #
        elif target[index] == "]":
            raise GdocSyntaxError("invalid syntax", (None, None, index, target))

        #
        # Non-quoted string
        #
        else:
            while index < length and (c := target[index]) != "]":

                if c == " ":  # string ended
                    break

                else:
                    name += c

                index += 1

        #
        # Wait closing ']'
        #
        while index < length:
            if target[index] == "]":
                break

            elif target[index] != " ":
                raise GdocSyntaxError("invalid syntax", (None, None, index, target))

            # else: target[i] == ' '
            index += 1

        if index == length:
            raise GdocSyntaxError(
                "EOS while scanning name string", (None, None, index - 1, target)
            )

        name = name.rstrip()

        return name, index + 1



r"""
PandocStr class
"""

from typing import Optional

_ALLOWED_TYPES_ = ("Str", "Space", "SoftBreak", "LineBreak")


class PandocStr:
    """
    Handles text strings in 'Str' inline elements and keep source mapping data.
    """

    def __init__(self, items=None, start: int = 0, stop: int = None):
        """Constructor
        @param items : [Str] | None
            PandocAST items to be added.
        @param start : int = 0
            start char pos in the str items.
        @param stop : int | None = None
            stop char pos in the str items.
        @return output : PandocStr
        """

        # self._items = []
        # self._text = ""
        # self._len = 0
        self._items, self._text, self._len = self._create_items_list(items, start, stop)

    def add_items(self, items=None, start: int = 0, stop: int = None):
        """Constructor
        @param items : [Str] | None
            PandocAST items to be added.
        @param start : int = 0
            start char pos in the str items.
        @param stop : int | None = None
            stop char pos in the str items.
        @return output : PandocStr
        """

        ## Create new items list
        # new_items = []
        # new_text = ""
        # new_len = 0
        new_items, new_text, new_len = self._create_items_list(items, start, stop)

        ## Join new list to existing lit.
        self._join_items(new_items, new_text, new_len)

    def get_items(self):
        """Constructor
        @return output : PandocStr
        """
        return self._items

    # def get_str(self, start: int = 0, stop: int = None):
    #     """Constructor
    #     @param start : int = 0
    #         start char pos in the str items.
    #     @param stop : int | None = None
    #         stop char pos in the str items.
    #     @return output : python str
    #     """
    #     return self._text[start:stop]

    def get_info(self, index: int = 0):
        """Constructor
        @param index : int = 0
            index of the target char in self._text
        @return (sourcepos : {path:str, line:int, col:int}, decoration, item)
        """
        _index = index
        sourcepos = None
        decoration = 0
        prev_ast_item = False

        for item in self._items:
            if _index >= item["len"]:
                _index -= item["len"]
                continue
            else:
                break

        if _index > item["len"]:
            # _index is longer than PandocStr._text.
            # Then refer to the last item and points the end of the item.
            prev_ast_item = True
            _index = 0

        # Pandoc AST elem type 'Str', 'Space', 'SoftBreak', 'LineBreak' don't have pos attr.
        # But their parents are 'Span' and have pos attr, when 'sourcepos'
        # extension is enabled.
        pos = item["_item"].get_parent().get_attr(("pos", "data-pos"))

        if (
            (pos is None)
            or (len(pos.split("@")) < 2)
            and (item["_item"].get_type() in ("Space", "SoftBreak"))
        ):
            # Currently(pandoc -v = 2.14.2),
            # If the type is SoftBreak, data-pos is not provided and is "".
            # Therefore, try to get the prev item and get its stop position.
            # The stop position points start point of the next(this) element.
            prev_ast_item = item["_item"].prev_item()
            if prev_ast_item is not None:
                pos = prev_ast_item.get_attr(("pos", "data-pos"))
                if (pos is None) or (len(pos.split("@")) < 2):
                    pos = prev_ast_item.get_parent().get_attr(("pos", "data-pos"))

        if (pos is not None) and (len(pos.split("@")) >= 2):
            # ".tmp/t.md@1:1-1:3"
            p = pos.split("@")
            _path = p[0]
            p = p[1].split("-")

            if prev_ast_item is False:
                p = p[0].split(":")
                _line = int(p[0])
                _col = int(p[1]) + item["start"] + _index
            else:
                p = p[1].split(":")
                _line = int(p[0])
                _col = int(p[1])

            sourcepos = {"path": _path, "line": _line, "col": _col}

        else:
            sourcepos = {"path": "[Source pos not found]", "line": 0, "col": 0}

        return sourcepos, decoration, item

    def __len__(self):
        """Constructor
        @return output : PandocStr
        """
        return self._len

    def __getitem__(self, index=0):
        """Constructor
        @param index : int | slice
        @return output : PandocStr
        """
        new_pandoc_str = None

        # 1. Setup indices
        if type(index) is int:
            start, stop, length = PandocStr._limit_index_to_range(index, self._len)

        elif type(index) is slice:
            start, stop, length = PandocStr._limit_slice_to_range(index, self._len)

        else:
            raise TypeError("PandocStr indices must be integers")

        # 2. Create empty pandocstr
        new_pandoc_str = PandocStr()

        # 3. Add all each items with start/stop info.
        if length > 0:
            # 3.1. move to start point
            for item in self._items:  # pragma: no branch: This line never complete.
                if start < item["len"]:
                    break
                else:
                    start -= item["len"]
                    continue

            # 3.2. add items for length
            for item in self._items[self._items.index(item) :]:  # pragma: no branch
                # This line never complete.
                _start = item["start"] + start

                if start + length < item["len"]:
                    _stop = _start + length
                else:
                    _stop = item["stop"]

                length -= item["len"] - start
                start = 0

                new_pandoc_str.add_items([item["_item"]], _start, _stop)

                if length <= 0:
                    break

        # 4. Return the new pandocstr
        return new_pandoc_str

    @classmethod
    def _limit_index_to_range(cls, index, length):
        if (index >= length) or (length + index < 0):  # index > 0  # index < 0
            raise IndexError("PandocStr index out of range")
        else:
            if index < 0:
                start = length + index
            else:
                start = index

        return start, (start + 1), 1

    @classmethod
    def _limit_slice_to_range(cls, index, length):
        # start
        if index.start is None:
            start = 0
        elif index.start >= 0:
            if index.start > length:
                start = length
            else:
                start = index.start
        else:
            start = length + index.start
            if start < 0:
                start = 0

        # stop
        if index.stop is None:
            stop = length
        elif index.stop >= 0:
            if index.stop > length:
                stop = length
            else:
                stop = index.stop
        else:
            stop = length + index.stop
            if stop < 0:
                stop = 0

        # length
        if stop >= start:
            _length = stop - start
        else:
            _length = 0

        return start, stop, _length

    def __str__(self) -> str:
        """Constructor
        @return str
        """
        return self._text

    def __contains__(self, x) -> bool:
        """Constructor
        @param x : str | PandocStr
        @return str
        """
        return str(x) in self._text

    def __eq__(self, o) -> bool:
        """Constructor
        @param o : str | PandocStr
        @return str
        """
        return str(o) == self._text

    def index(self, value, start: int = 0, stop: int = None):
        """Constructor
        @param value : str | PandocStr
        @param start : int = 0
        @param stop : int | None = None
        @return index : int
        """
        return self._text.index(str(value), start, stop)

    def count(self, value) -> int:
        """Constructor
        @param value : str | PandocStr
        @return str
        """
        return self._text.count(str(value))

    def __add__(self, s):
        """Constructor
        @param s : PandocStr | str
        @return PandocStr | str
        """
        new_pandoc_str = None

        if isinstance(s, PandocStr):
            new_pandoc_str = self[:]
            opr_str = s[:]
            new_pandoc_str._join_items(opr_str._items, opr_str._text, opr_str._len)

        elif type(s) is str:
            new_pandoc_str = self._text + s

        else:
            raise TypeError(
                'can only concatenate PandocStr or str (not "'
                + s.__class__.__name__
                + '") to PandocStr'
            )

        return new_pandoc_str

    def __radd__(self, s):
        """Constructor
        @param s : PandocStr | str
        @return PandocStr | str
        """
        new_pandoc_str = None

        if isinstance(s, PandocStr):
            new_pandoc_str = s[:]
            opr_str = self[:]
            new_pandoc_str._join_items(opr_str._items, opr_str._text, opr_str._len)

        elif type(s) is str:
            new_pandoc_str = s + self._text

        else:
            raise TypeError(
                'can only concatenate PandocStr or str (not "'
                + s.__class__.__name__
                + '") to PandocStr'
            )

        return new_pandoc_str

    def __iadd__(self, s):
        """Constructor
        @param s : PandocStr | str
        @return PandocStr | str
        """

        if isinstance(s, PandocStr):
            opr_str = s[:]
            self._join_items(opr_str._items, opr_str._text, opr_str._len)

        else:
            raise TypeError(
                'can only concatenate PandocStr (not "'
                + s.__class__.__name__
                + '") to PandocStr'
            )

        return self

    def startswith(self, *args, **kwargs):

        return self._text.startswith(*args, **kwargs)

    def endswith(self, *args, **kwargs):

        return self._text.endswith(*args, **kwargs)

    def strip(self, *args, **kwargs) -> "PandocStr":
        word = self._text.strip(*args, **kwargs)
        start = self.index(word)
        stop = start + len(word)
        return self[start:stop]

    def lstrip(self, __chars: Optional[str] = None) -> "PandocStr":
        word = self._text.lstrip(__chars)
        start = self.index(word)
        stop = start + len(word)
        return self[start:stop]

    def rstrip(self, __chars: Optional[str] = None) -> "PandocStr":
        word = self._text.rstrip(__chars)
        start = self.index(word)
        stop = start + len(word)
        return self[start:stop]

    def find(self, *args, **kwargs) -> int:
        return self._text.find(*args, **kwargs)

    def isspace(self) -> bool:
        return self._text.isspace()

    def _create_items_list(self, items=None, start: int = 0, stop: int = None):
        """Constructor
        @param items : [Str] | None
            PandocAST items to be added.
        @param start : int = 0
            start char pos in the str items.
        @param stop : int | None = None
            stop char pos in the str items.
        @return output : PandocStr
        """
        ##
        ## Create new items list
        ##
        new_items = []
        new_text = ""
        new_len = 0
        _stop = stop  # storing for error reporting

        if items is None or len(items) == 0:
            return (new_items, new_text, new_len)

        total_lenght = 0
        for item in items:
            if item.get_type() in _ALLOWED_TYPES_:
                new_items.append(
                    {
                        "_item": item,
                        "start": 0,
                        "stop": len(item.text),
                        "text": item.text,
                        "len": len(item.text),
                        "decoration": 0,
                    }
                )
                total_lenght += len(item.text)

            else:
                # raise error
                raise TypeError("Invalid item type(" + item.get_type() + ")")

        # Set stop as plus val(0 - total_length)
        if stop is None:
            stop = total_lenght
        elif stop < 0:
            stop = total_lenght + stop

        trailing_str_len = total_lenght - stop

        # Check invalid start and stop
        if total_lenght > 0:
            if (start < 0) or (start > total_lenght):
                raise IndexError("Out of range specifier: start = " + str(start))

            if (stop < 0) or (stop > total_lenght):
                raise IndexError("Out of range specifier: stop = " + str(_stop))

        new_len = stop - start  # if start pos == stop pos: len = 0
        if new_len < 0:
            # raise error
            raise ValueError("Invalid range specification")

        # Check `start` value
        while start > 0:
            length = new_items[0]["len"]

            if length <= start:
                del new_items[0]
                start -= length

            else:
                new_items[0]["start"] = start
                new_items[0]["len"] -= start
                break

        # Check `stop` value
        while trailing_str_len > 0:
            length = len(new_items[-1]["_item"].text)

            if length < trailing_str_len:
                del new_items[-1]
                trailing_str_len -= length

            else:
                new_items[-1]["stop"] = length - trailing_str_len
                new_items[-1]["len"] -= trailing_str_len
                break

        # remove empty items and text out of range.
        for item in new_items:
            if item["len"] == 0:
                del new_items[new_items.index(item)]
                continue

            # remove text out of range.
            elif (item["start"] > 0) or (item["stop"] < len(new_items[-1]["_item"].text)):
                item["text"] = item["text"][item["start"] : item["stop"]]

            new_text += item["text"]

        return (new_items, new_text, new_len)

    def _join_items(self, new_items, new_text, new_len):
        """Constructor
        @param items : [Str] | None
            PandocAST items to be added.
        @param start : int = 0
            start char pos in the str items.
        @param stop : int | None = None
            stop char pos in the str items.
        @return output : PandocStr
        """

        ##
        ## Join new list to existing lit.
        ##
        if (
            ((self._len > 0) and (new_len > 0))
            and (self._items[-1]["_item"] is new_items[0]["_item"])
            and (self._items[-1]["stop"] == new_items[0]["start"])
        ):

            # Merge them
            # "_item": item,            <- same
            # "start" : 0,              <- same
            # "stop": len(item.text),   <- replace
            # "text": item.text,        <- add
            # "len": len(item.text),    <- add
            # "decoration": 0           <- same
            self._items[-1]["stop"] = new_items[0]["stop"]
            self._items[-1]["len"] += new_items[0]["len"]
            self._items[-1]["text"] += new_items[0]["text"]
            del new_items[0]

        self._items += new_items
        self._text += new_text
        self._len += new_len



from . import block, requirement

exports = {
    "Table": {
        "requirement": {"constructor": requirement.Requirement},
        "reqt": {"constructor": requirement.Requirement},
        "block": {"constructor": block.Block},
        "blk": {"constructor": block.Block},
    }
}



r"""
Block class
"""

from .element import Element


class Block(Element):
    """
    Base class of PandocAST Element handler.
    """

    def __init__(self, pan_elem, elem_type, type_def):
        """Constructor
        @param pan_elem(Dict)
            PandocAST Element
        @param elem_type(Str)
            Element type
        @param type_def(Dict)
            Element structur data
        """
        super().__init__(pan_elem, elem_type, type_def)



"""
errorhandler.py: ErrorHandler class
"""


from typing import Union, cast


class ErrorReport:
    """
    Error handling
    """

    def __init__(self, cont: bool = False) -> None:
        self._errordata: list[Exception | "ErrorReport"] = []
        self._exit: bool = not cont

    def submit(self, err: Union[Exception, "ErrorReport", None] = None) -> bool:
        if (err is not None) and (err is not self):
            self._errordata.append(err)

        return self._exit

    def new_subreport(self) -> "ErrorReport":
        return self.__class__(not self._exit)

    def haserror(self) -> bool:
        return len(self._errordata) > 0

    def get_errors(self) -> list[Exception]:
        errors: list[Exception] = []

        for err in self._errordata:
            if type(err) is ErrorReport:
                errors += err.get_errors()
            else:
                err = cast(Exception, err)
                errors.append(err)

        return errors

    def dump(self) -> str:
        dumpstrings: list[str] = []
        errstr: str

        errors: list[Exception] = self.get_errors()
        for err in errors:
            if isinstance(err, SyntaxError):
                errstr = f"{err.filename}:{err.lineno}:{err.offset} "
                if err.end_offset is not None:
                    errstr += f"- {err.end_lineno}:{err.end_offset} "
                errstr += f"{type(err).__name__}: {str(err)}\n"
                errstr += f"> {err.text}"
                if err.offset is not None:
                    errstr += "\n>" + (" " * err.offset) + "^"
            else:
                errstr = str(err)

            dumpstrings.append(errstr)

        return "\n".join(dumpstrings)



"""
code.py: `Code` inline element class
"""

from typing import cast

from gdoc.lib.pandocastobject.pandocast import PandocAst, PandocElement

from .text import Text


class Code(Text):
    """ """

    def __init__(self, item: PandocElement | str):
        element: PandocElement
        if type(item) is str:
            element = PandocAst.create_element({"t": "Code", "c": [["", [], []], item]})
        else:
            element = cast(PandocElement, item)

        if element.get_type() != "Code":
            raise RuntimeError()

        self.element = element

    def get_str(self):
        return f"`{self.element.get_content()}`"

    def get_text(self):
        return self.element.get_content()



"""
Command line interface module
"""
import argparse
import importlib
import logging
import os

from . import _CONFIG


def main():
    """
    * Command line interface
    """
    parser = argparse.ArgumentParser(prog=__package__)
    parser.add_argument("-v", "--version", action="store_true", help="show version")

    common = argparse.ArgumentParser(add_help=False)
    common.add_argument(
        "--debug", nargs="*", help="module path to output logger.debug()."
    )

    subparsers = parser.add_subparsers(title="commands")

    here = os.path.dirname(__file__)
    files = os.listdir(os.path.join(here, _CONFIG["app_path"]))
    for file in files:
        if os.path.isdir(os.path.join(here, _CONFIG["app_path"], file)) and (
            file != "__pycache__"
        ):
            importlib.import_module(
                __package__ + "." + _CONFIG["app_path"] + "." + file
            ).setup(subparsers, file, common)

    args = parser.parse_args()

    if hasattr(args, "func"):
        if hasattr(args, "debug") and args.debug is not None:
            if len(args.debug) > 0:
                logging.basicConfig()
                for module in args.debug:
                    module = ".".join(module.split("/"))
                    logging.getLogger(module).setLevel(level=logging.DEBUG)
            else:
                logging.basicConfig(level=logging.DEBUG)

        args.func(args)

    elif args.version:
        print(_CONFIG["version"])

    else:
        parser.print_help()


if __name__ == "__main__":
    main()



DEFAULTS: dict = {
    "pandocast": {
        "types": {
            "ignore": ["Div", "Span"],
            "remove": ["Strikeout"],
            "decorator": [
                "Span",
                "Emph",
                "Underline",
                "Strong",
                "Superscript",
                "Subscript",
                "SmallCaps",
                "Link",
            ],
            "plaintext": ["Str", "Space", "SoftBreak"],
            "textblock": ["Plain", "Para", "LineBlock", "Header"],
        },
        "altchar": {"Code": "`", "Math": "$", "Quoted": '"', "RawInline": ""},
    },
    "textblock": {"types": ["Plain", "Para", "LineBlock", "Header"]},
}



"""
tag.py: tag class
"""
from typing import NamedTuple, Optional

from gdoc.lib.gdoc import String, TextString

from .textblock import TextBlock


class BlockTagInfo(NamedTuple):
    textblock: TextBlock
    preceding_lines: list[TextString]
    following_lines: list[TextString]
    preceding_text: Optional[TextString]
    following_text: Optional[TextString]


class BlockTag(TextString):
    """
    BlockTag class
    """

    _class_info: tuple[String | None, String | None, String | None]
    category: String | None
    type: String | None
    isref: String | None
    class_args: list[TextString]
    class_kwargs: list[tuple[TextString, TextString]]
    tag_text: TextString
    tag_info: BlockTagInfo | None = None

    def __init__(
        self,
        class_info: tuple[String | None, String | None, String | None],
        class_args: list[TextString],
        class_kwargs: list[tuple[TextString, TextString]],
        tag_text: TextString,
        tag_info: BlockTagInfo | None = None,
    ):
        super().__init__(tag_text)

        self._class_info = class_info
        self.category, self.type, self.isref = class_info
        self.class_args = class_args
        self.class_kwargs = class_kwargs
        self.tag_text = tag_text
        self.tag_info = tag_info

    def get_class_arguments(
        self,
    ) -> tuple[
        tuple[String | None, String | None, String | None],
        list[TextString],
        list[tuple[TextString, TextString]],
    ]:
        """
        _summary_

        @return _type_ : _description_
        """
        return self._class_info, self.class_args, self.class_kwargs



"""
documentparser.py: parse_Document function
"""
from gdoc.lib.gdoc import Document
from gdoc.lib.gobj.types import BaseObject
from gdoc.util import Result

from ...util.errorreport import ErrorReport
from .sectionparser import parse_Section


def parse_Document(
    document: Document, gobj: BaseObject, opts: dict, erpt: ErrorReport
) -> Result[BaseObject, ErrorReport]:
    """
    parse gdoc.Document

    Add Here:
    - read meta data and set opts based on it.
    """

    return parse_Section(document, gobj, opts, erpt)



"""
blocktagparser.py: parse_BlockTag function
"""
from typing import Optional, Union, cast

from gdoc.lib.gdoc import String, Text, TextString
from gdoc.lib.gdoc.blocktag import BlockTag
from gdoc.util import Err, Ok, Result
from gdoc.util.errorreport import ErrorReport

from ..fsm import State, StateMachine
from ..textblock.texttokenizer import TextTokenizer
from .objecttaginfoparser import ObjectTagInfo, parse_ObjectTagInfo


def parse_BlockTag(
    tokenized_textstr: TextString, start: int, opts: dict, erpt: ErrorReport
) -> Result[tuple[TextString, Optional[int]], ErrorReport]:
    """
    _summary_

    @param tokenized_textstr (TextString) : _description_
    @param start (int) : _description_
    @param opts (dict) : _description_
    @param erpt (ErrorReport) : _description_

    @return Result[tuple[TextString, Optional[int], ErrorReport]] : _description_
    """
    textstr: TextString = tokenized_textstr
    tag_start: Optional[int] = None
    tagpos: Optional[slice] = None
    tagstr: Optional[TextString] = None
    blocktag: Optional[BlockTag] = None

    tagpos, tagstr = detect_BlockTag(tokenized_textstr, start)

    if tagpos is not None:
        tagstr = cast(TextString, tagstr)

        tokens: TextString = tagstr[1:-1]
        # remove the first "[@" and the last "]"
        # TODO: Replace with removeprefix("[@") and removesuffix("]")

        taginfo: ObjectTagInfo
        taginfo, e = parse_ObjectTagInfo(tokens, opts, erpt)
        if e:
            erpt.submit(e)
            return Err(erpt)

        class_info: tuple[String | None, String | None, String | None]
        class_args: list[TextString]
        class_kwargs: list[tuple[TextString, TextString]]

        class_info, class_args, class_kwargs = taginfo
        blocktag = BlockTag(class_info, class_args, class_kwargs, tagstr)
        textstr = tokenized_textstr[:]
        textstr[tagpos] = [blocktag]
        tag_start = tagpos.start

    return Ok((textstr, tag_start))


def detect_BlockTag(
    tokenized_textstr: TextString, start: int
) -> tuple[Optional[slice], Optional[TextString]]:
    result: tuple[Optional[slice], Optional[TextString]]
    tagpos: list[int]
    tagstr: Optional[TextString]

    detector: StateMachine = BlockTagDetector()
    start_pos: int = start

    while True:
        detector.start().on_entry()

        for i, token in enumerate(tokenized_textstr[start_pos:]):
            if detector.on_event((i, token)) is None:
                break

        tagpos, tagstr = detector.on_exit()

        if tagpos[0] < 0:
            # Opening str "[@" NOT FOUND
            result = (None, None)
            break

        elif tagpos[1] < 0:
            # "[@" has been found, but NOT END correctly with "]".
            # Retry from the location indicated with tag_start.
            start_pos += tagpos[0] + 2  # 2 = len('[@')

        else:
            # Tag detected
            result = (slice(start_pos + tagpos[0], start_pos + tagpos[1]), tagstr)
            break

        detector.stop()

    return result


#
# BlockTag detector
#
class BlockTagDetector(StateMachine):
    """ """

    def __init__(self, name: str = None) -> None:
        super().__init__(name)
        self.tagstr: Optional[TextString] = None
        self.tagpos: Optional[list[int]] = None

        self.add_state(_Open("Opening"), "Character")
        self.add_state(_Char("Character"), None)
        self.add_state(_String("String"), "Character")

    def start(self, param=None):
        self.tagstr = TextString()
        self.tagpos = [-1, -1]
        return super().start((self.tagstr, self.tagpos))

    def on_entry(self, _=None):
        return super().on_entry(None)  # None for reset substatuses

    def on_exit(self):
        super().on_exit()
        return self.tagpos, self.tagstr

    def stop(self):
        self.tagstr = None
        self.tagpos = None
        return super().stop()


class _Open(State):
    """ """

    def __init__(self, name: str = None) -> None:
        super().__init__(name)
        self.tagstr: Optional[TextString] = None
        self.tagpos: Optional[list[int]] = None
        self._prev: Optional[String] = None
        self._start: int = -1

    def start(self, taginfo):
        self.tagstr, self.tagpos = taginfo

    def on_entry(self, _=None):
        self._prev = None
        self._start = -1
        return self

    def on_event(self, event: tuple[int, Text]):
        next: Optional[State] = self
        index, token = event

        if self._prev is None:
            if isinstance(token, String) and (token == "["):
                self._prev = token
                self._start = index

        elif isinstance(token, String) and (token == "@"):
            cast(TextString, self.tagstr).append(self._prev + token)
            cast(list[int], self.tagpos)[0] = self._start
            next = None

        else:
            self._prev = None
            self._start = -1

        return next

    def stop(self):
        self.tagstr = None
        self.tagpos = None


class _Char(State):
    """ """

    def __init__(self, name: str = None) -> None:
        super().__init__(name)
        self.tagstr: Optional[TextString] = None
        self.tagpos: Optional[list[int]] = None
        self._bcount: int = 0
        self._word: Optional[String] = None

    def start(self, taginfo):
        self.tagstr, self.tagpos = taginfo
        self._bcount = 0

    def on_entry(self, _=None):
        next: Optional[State] = self

        self._word = String()  # Empty string

        return next

    def on_event(self, event):
        next: Union[State, None, tuple[str, tuple[int, Text]]] = self
        index, token = event

        if isinstance(token, String) and (token == '"'):
            next = ("String", event)

        elif TextTokenizer.is_word(token):
            self._word += token

        else:
            self._flush_word_buff()
            self.tagstr.append(token)

            if isinstance(token, String) and (token == "["):
                self._bcount += 1

            elif isinstance(token, String) and (token == "]"):
                if self._bcount > 0:
                    self._bcount -= 1
                else:
                    # Tag Detection Success
                    cast(list[int], self.tagpos)[1] = index + 1
                    next = None

        return next

    def stop(self):
        self.tagstr = None
        self.tagpos = None

    def _flush_word_buff(self):
        if len(self._word) > 0:
            # Flush buffer
            self.tagstr.append(self._word)
            self._word = String()  # Empty string


class _String(State):
    """ """

    def __init__(self, name: str = None) -> None:
        super().__init__(name)
        self.tagstr: Optional[TextString] = None
        self.tagpos: Optional[list[int]] = None
        self.quoted: Optional[TextString] = None
        self.escape: bool = False

    def start(self, taginfo):
        self.tagstr, self.tagpos = taginfo

    def on_entry(self, event):
        next: Union[State, None, tuple[str, tuple[int, Text]]] = self
        _, token = event

        self.quoted = TextString([token])  # Always it's '"'.
        self.escape = None

        return next

    def on_event(self, event):
        next: Union[State, None, tuple[str, tuple[int, Text]]] = self
        _, token = event

        self.quoted.append(token)

        if self.escape:
            self.escape = False

        elif isinstance(token, String) and (token == "\\"):
            self.escape = True

        elif isinstance(token, String) and (token == '"'):
            self.tagstr.append(self.quoted)
            self.quoted = None
            next = None

        return next

    def stop(self):
        self.tagstr = None
        self.tagpos = None



r"""
Document class
"""
from . import CATEGORY_INFO
from .baseobject import BaseObject
from .category import Category


class Document(BaseObject):
    """ """

    Category(CATEGORY_INFO)

    def __init__(self, id, name=None):
        """ """
        super().__init__("DOCUMENT", id, name=name)



"""
textstringparser.py: parse_Line function
"""
from typing import Optional

from gdoc.lib.gdoc import TextString
from gdoc.util import Err, Ok, Result

from ....util.errorreport import ErrorReport
from ..tag.blocktagparser import parse_BlockTag
from .texttokenizer import tokenize_textstring


def parse_Line(
    textstr: TextString, opts: dict, erpt: ErrorReport
) -> Result[TextString, ErrorReport]:
    """
    _summary_

    @param textstr (TextString) : Tokenized TextString
    @param opts (dict) : _description_
    @param erpt (ErrorReport) : _description_

    @return Result[TextString, ErrorReport] : _description_
    """
    tokenized_textstr: TextString
    tag_pos: Optional[int]

    # Tokenize
    tokenized_textstr = tokenize_textstring(textstr)

    # Parse BlockTag(s) in TextString
    tag_pos = -1
    while tag_pos is not None:  # if None, parsed to the EOL.
        parseresults: Optional[tuple[TextString, Optional[int]]]
        parseresults, e = parse_BlockTag(tokenized_textstr, tag_pos + 1, opts, erpt)

        if e and erpt.submit(e):
            return Err(erpt)

        if parseresults is None:
            break

        tokenized_textstr, tag_pos = parseresults
        # a part of tokenized_text was replaced with a BlockTag.

    # Detect InlineTags
    # tag_index = -1
    # while tag_index is not None:
    #     tokenized_textstr, tag_index = parse_InlineTag(tokenized_textstr, tag_index + 1)
    #     # replaced a part of tokenized_text to with a InlineTag.

    return Ok(tokenized_textstr)



"""
baseobject.py: BaseObject class
"""
from typing import Any, NamedTuple, final

from gdoc.lib.gdoc import String, TextString
from gdoc.lib.gdoccompiler.gdexception import *

from ..gdobject import GdObject
from ..gdsymbol import GdSymbol
from ..gdsymboltable import GdSymbolTable
from .category import Category


class ClassInfo(NamedTuple):
    category: str | String | None
    type: str | String | None
    isref: bool | String | None


class BaseObject(GdObject):
    """
    BaseObject class
    """

    class_category: str
    class_type: str
    class_version: str
    class_isref: bool

    def __init__(
        self,
        typename: GdSymbolTable.Type | str,
        id,
        scope="+",
        name=None,
        tags=[],
        ref=None,
        type_args={},
    ):
        if type(typename) is GdSymbolTable.Type:
            _type = typename
            typename = typename.name

        else:
            if ref is None:
                _type = GdSymbolTable.Type.OBJECT
            else:
                _type = GdSymbolTable.Type.REFERENCE

        if scope is None:
            scope = "+"
        super().__init__(id, scope=scope, name=name, tags=tags, _type=_type)

        cat = self.__class__.get_category()
        if type(cat) is Category:
            self.class_category = cat.name
            self.class_version = cat.version
        else:
            self.class_category = ""
            self.class_version = ""

        self.class_type = typename
        self[""].update(
            {
                "class": {
                    "category": self.class_category,
                    "type": self.class_type,
                    "version": self.class_version,
                }
            }
        )
        if ref is not None:
            self.class_isref = True
            self[""]["class"]["ref"] = {"object_path": ref}
        else:
            self.class_isref = False

        self.update(type_args)

    @final
    def create_object(
        self,
        class_info: tuple[String | str | None, String | str | None, String | str | None],
        class_args: list[TextString],
        class_kwargs: list[tuple[TextString, TextString]],
        tag_opts: dict,
        tag: Any,
    ) -> "BaseObject":
        """
        _summary_

        @param class_info (ClassInfo) : _description_
        @param class_args (list[TextString]) : _description_
        @param class_kwargs (list[tuple[TextString, TextString]]) : _description_
        @param tag_opts (dict, optional) : _description_. Defaults to {}.

        @exception GdocSyntaxError : _description_
        @exception GdocRuntimeError : _description_
        @exception GdocRuntimeError : _description_
        @exception GdocTypeError : _description_

        @return BaseObject : _description_
        """
        class_cat = str(class_info[0]) if class_info[0] is not None else None
        class_type = str(class_info[1]) if class_info[1] is not None else ""
        constructor = None

        obj = self
        while obj is not None:
            class_type, constructor = obj.__get_constructor(class_cat, class_type)
            if constructor is not None:
                break

            obj = obj.get_parent()

        if constructor is not None:
            child = constructor.create(
                class_info, class_args, class_kwargs, tag_opts, parent_obj=self
            )

        else:
            raise GdocTypeError("Class not found")

        return child

    def __get_constructor(self, class_cat: str = None, class_type: str = ""):
        """ """
        constructor = None
        class_name = None

        if class_cat in (None, self.class_category):
            cat = self.__class__.get_category()

        if type(cat) is Category:
            class_name, constructor = cat.get_type(class_type, self.class_type)

        return class_name, constructor

    @classmethod
    def create(
        cls,
        class_info: tuple[String | str | None, String | str | None, String | str | None],
        class_args: list[TextString],
        class_kwargs: list[tuple[TextString, TextString]],
        tag_opts: dict,
        parent_obj: "BaseObject",
    ) -> "BaseObject":
        id = None
        tags = []
        isref = class_info[2]  # isref
        ref = None
        scope, symbol, args = _get_symbol(class_args)

        if symbol is not None:
            if type(symbol) is str:
                symbol = GdSymbol(symbol)
            else:
                symbol = GdSymbol(symbol.get_text())

            if not symbol.is_id():
                raise GdocSyntaxError("Invalid id")

            symbols = symbol.get_symbols()
            id = symbols[-1]

            if isref:
                ref = symbols
            else:
                ref = None
                # assert here if symbols[0:-1] are valid
                symbols.pop()
                p = parent_obj.get_parent()
                while len(symbols) > 0:
                    s = symbols.pop()
                    if s.startwith("*"):  # name
                        if p.name != s[1:]:
                            raise GdocRuntimeError(
                                "The explicit parent Name is incorrect."
                            )
                    else:  # id
                        if p.id != s:
                            raise GdocRuntimeError("The explicit parent ID is incorrect.")

            tags = symbol.get_tags()

        child = cls(
            str(class_info[1]),  # type
            id,
            scope=scope,
            name=tag_opts.get("name"),
            tags=tags,
            ref=ref,
            type_args=tag_opts,
        )

        if parent_obj:
            parent_obj.add_child(child)

        return child


def _get_symbol(class_args):
    scope = None
    symbol = None
    args = []

    idx = 0
    c = len(class_args)
    if c > 0:
        if class_args[idx].get_text() in ("+", "-"):
            scope = class_args[idx]
            if c < 2:
                raise GdocSyntaxError()
                # Symbol should follow
            idx += 1

        # class_args[idx] should be symbol
        symbol = class_args[idx]
        idx += 1

        if symbol.startswith(("+", "-")):
            # TODO: check that the first element is a String. If not,
            # the following deletion of the first character does not work.
            if scope is None:
                scope = symbol[0][0]
                symbol[0] = symbol[0][1:]

            else:
                raise GdocSyntaxError()
                # Scope is duplecated

        args = class_args[idx:]

    return scope, symbol, args



#
# Ghost document Object Model
#
# gdoc '@'記号によるタグ付けを解釈し、以下を実現する
# 1. Packaging
#    - Namespaceを構成し、名前解決手段を提供する
#    - import/accessに対応し、他パッケージへのアクセス手段を提供する
# 2. Object抽出
#    - タグを解釈し、オブジェクトの検出を行う
#    - 検出したオブジェクトのクラスを判断し、クラスに応じたオブジェクト構造の解析を行う
# 3. 基本データ型解釈
#    - ID、文字列、数値、真偽値などの基本型を、クラスの期待とタグ付けに従って解釈する
# 4. アクセスライブラリ
#    - 上位のプラグインがデータを取り扱いやすいよう、アクセスライブラリを提供する
#
# 以上から、Gdoc Objectを出力する。

# from . import gdast
from gdoc.lib.pandocast import pandocast as gdast

from .symboltable import Scope, Symbol, SymbolTable


#
# Primitive types
#
class Element:
    def __init__(self, parent=None, tag=None) -> None:
        super().__init__()

        self.parent = parent
        self.source = {}
        if tag is None:
            self.id = ""
            self.name = ""
            self.plugin = ""
            self.types = []
            self.option = {}
        else:
            self.id = tag.id
            self.name = tag.name
            self.plugin = tag.plugin
            self.types = tag.types[:]
            self.option = tag.option.copy()


class Object(Element):
    def __init__(self, parent=None, tag=None) -> None:
        super().__init__(parent, tag)

        self.content = None

    def getItems():
        # should raise
        return "NOT YET IMPLIMENTED"

    def dump(self):
        return self.content


class Package(Element):
    def __init__(self, parent, level=0, tag=None) -> None:
        super().__init__(parent, tag)

        self._level = level
        self.children = []

        parentTable = None if parent is None else parent.symbolTable
        self.symbolTable = SymbolTable(self, parentTable)
        if parentTable is not None:
            parentTable.addItem(tag.id, tag.name, Package, self.symbolTable)

    def dump(self):
        data = [{}, []]
        data[0]["type"] = "package"
        data[0]["id"] = self.id
        data[0]["name"] = self.name
        data[0]["plugin"] = self.plugin
        data[0]["types"] = self.types
        data[0]["option"] = self.option

        for child in self.children:
            data[1].append(child.dump())

        return data

    def resolve(self, name):
        return self.symbolTable.resolve(name)


class PackageManager(Package):
    def __init__(self, parent, level=0, tag=None) -> None:
        super().__init__(parent, level, tag)

        self._current = self
        self._contextStack = [self]
        self.children = []

    def _addSection(self, level, tag):
        while (level <= self._current._level) and (
            self._contextStack[-1] != self._current
        ):
            self._current = self._current.parent

        # open next package
        package = Package(self._current, level, tag)
        self._current.children.append(package)
        self._current = package

    def _addObject(self, element):
        self._current.children.append(element)
        items = element.getItems(self._current.symbolTable)
        for item in items:
            self._current.symbolTable.addItem(**item)

    def _pushContext(self):
        self._contextStack.append(self._current)
        pass

    def _popContext(self):
        while self._contextStack[-1] != self._current:
            self._current = self._current.parent
        del self._contextStack[-1]


class BlockTag:
    def __init__(self, line) -> None:
        self.line = line
        self._istagged = False
        self._class = ""
        self.id = ""
        self.name = ""
        self.plugin = None
        self.types = []
        self.option = {}

        self._parse()

    def _parse(self):
        if (hstart := self.line.find("[@")) < 0:
            return

        if (hend := self.line.find("]", hstart + 2)) < 0:  # 2 = len('[@')
            return

        self._istagged = True

        tagHeader = self.line[
            hstart + 1 : hend
        ]  # [ + '@class id..' + ] <- picking this center str.
        tagBody = self.line[hend + 1 :]

        words = tagHeader.split()
        if words[0] != "@":
            self._class = words[0][1:]  # removing '@' at the top.
            classes = self._class.split(":")

            if len(classes) > 2:
                self.plugin = ["Syntax Error"]
                return

            elif len(classes) == 2:
                self.plugin = classes[0]
                del classes[0]

            self.types = classes[0].split(".")
            if (len(self.types) > 1) and ("" in self.types):
                self.types = ["Syntax Error"]

        if len(words) > 1:
            self.id = words[1]
            # タグオプション記法をそのうち実装する

        self.name = tagBody.strip()
        # 行末オプション記法をそのうち実装する

    def istagged(self) -> bool:
        return self._istagged

    def getParser(self):
        return False


class GdocObjectModel(PackageManager):
    def __init__(self, gdoc, types) -> None:

        # Create ROOT package.
        # - ROOT package should contain source.filename etc. for import/access.
        super().__init__(None)

        self.objectTypes = types

        # Step.1: Syntax analysis
        # Call parser with arg, package and gdoc element.
        self.parseBlocks(gdoc.getFirstChild())

    def parseBlocks(self, block):

        while block is not None:
            # 概要：
            # 1. 与えられたブロック要素を起点にブロック終端（next()==None）まで走査を行う。
            # 2. Headerセクションの場合は、単にPackageStackを積み上げる。
            # 3. Listセクションによりブロックリストが捜査対象になった場合は、再帰呼び出しを行う。
            # 4. オブジェクトを見つけた場合には、現在地にオブジェクトを登録する。

            # element type check
            if not isinstance(block, gdast.Block):
                return (
                    __class__
                    + "->Error: Invalid element type(should be instance of Block)"
                )

            line = block.getFirstLine()
            if isinstance(block, gdast.Table):
                line = "[" + line + "]"
            tag = BlockTag(line)

            # Header:
            if block.type == "Header":
                if tag.istagged():
                    self._addSection(block.getProp("Level"), tag)

            # List:
            elif (block.type in ["OrderedList", "BulletList"]) and not (
                (block.getFirstChild().getFirstChild().type in ["Plain", "Para"])
                and tag.istagged()
            ):
                listitem = block.getFirstChild()
                while listitem is not None:
                    self._pushContext()
                    self.parseBlocks(listitem.getFirstChild())
                    self._popContext()
                    listitem = listitem.next()

            # Object:
            elif tag.istagged():
                constructor = self.objectTypes.getConstructor(
                    self._current, block.type, tag.plugin, tag.types
                )
                if callable(constructor):
                    obj = constructor(block, tag)
                    self._addObject(obj)

            else:
                # blocks without tag
                pass

            block = block.next()

    def walk(self, action, post_action=None):
        self._walk(action, post_action, self)

    def _walk(self, action, post_action, element):

        action(element)

        if hasattr(element, "children"):
            for child in element.children[:]:
                self._walk(action, post_action, child)

        if post_action is not None:
            post_action(element)



r"""
GdObject class
"""
from gdoc.lib.gdoccompiler.gdexception import *

from .gdsymboltable import GdSymbolTable


class GdObject(GdSymbolTable):
    """
    ;
    """

    __category_module = None

    @classmethod
    def set_category(cls, module):
        cls.__category_module = module

    @classmethod
    def get_category(cls):
        return cls.__category_module

    def __init__(self, *args, **kwargs):
        """Constructs GdObject.
        @param id : str | PandocStr
        """
        super().__init__(*args, **kwargs)

        self.__properties = {
            "": {
                "id": self.id,
                "scope": self.scope,
                "name": self.name,
                "tags": self.tags[:],
            }
        }

    def set_prop(self, key, value):
        """split symbol string to ids or names and tags.
        @param symbol : str | PandocStr
        @return (list(str), list(str))
        """
        if type(key) is list:
            keys = key
        else:
            keys = [key]

        for k in keys:
            if type(k) is not str:
                raise GdocTypeError("invalid key type")

            elif k == "":
                raise GdocKeyError('invalid key ""')

        dst = self.__properties
        for k in keys[:-1]:
            v = dst.get(k)

            if v is None:
                dst[k] = {}
                dst = dst[k]

            elif type(v) is dict:
                dst = v

            else:
                dst[k] = {"": v}
                dst = dst[k]

        k = keys[-1]
        v = dst.get(k)

        if v is None:
            dst[k] = value

        else:
            if type(v) is dict:
                dst = v
                k = ""
                v = dst.get(k)

            if v is None:
                dst[k] = value

            elif type(v) is list:
                if type(value) is list:
                    dst[k] = v + value
                else:
                    v.append(value)

            else:
                if type(value) is list:
                    dst[k] = [v] + value
                else:
                    dst[k] = [v, value]

    def get_prop(self, key):
        """split symbol string to ids or names and tags.
        @param symbol : str | PandocStr
        @return (list(str), list(str))
        """
        if type(key) is list:
            keys = key
        else:
            keys = [key]

        for k in keys:
            if type(k) is not str:
                raise GdocTypeError("invalid key type")

            elif k == "":
                raise GdocKeyError('invalid key ""')

        dst = self.__properties
        for k in keys[:-1]:
            v = dst.get(k)

            if type(v) is dict:
                dst = v
            else:
                return None

        k = keys[-1]
        if k not in dst:
            return None

        v = dst[k]

        if type(v) is dict:
            v = v.get("")
            if v is None:
                return []

        result = None
        if type(v) is list:
            result = v
        else:
            result = [v]

        return result

    def __getitem__(self, *args, **kwargs):
        return self.__properties.__getitem__(*args, **kwargs)

    def __iter__(self, *args, **kwargs):
        return self.__properties.__iter__(*args, **kwargs)

    def __len__(self, *args, **kwargs):
        return self.__properties.__len__(*args, **kwargs)

    def __contains__(self, *args, **kwargs):
        return self.__properties.__contains__(*args, **kwargs)

    def __eq__(self, *args, **kwargs):
        return self.__properties.__eq__(*args, **kwargs)

    def __ne__(self, *args, **kwargs):
        return self.__properties.__ne__(*args, **kwargs)

    def keys(self, *args, **kwargs):
        return self.__properties.keys(*args, **kwargs)

    def items(self, *args, **kwargs):
        return self.__properties.items(*args, **kwargs)

    def values(self, *args, **kwargs):
        return self.__properties.values(*args, **kwargs)

    def get(self, *args, **kwargs):
        return self.__properties.get(*args, **kwargs)

    def update(self, *args, **kwargs):
        return self.__properties.update(*args, **kwargs)



# Gdoc object types

from ..plugin import sysml


class Plugins:
    def __init__(self, pluginpath=None) -> None:
        self.typetable = _TypeTable
        self.typetable["sysml"] = sysml.exports
        pass

    # self, block.type, tag.plugin, tag.types
    def getConstructor(self, package, blocktype, plugin, types):
        constructor = None
        if plugin is None:
            plugin = ""
            pkg = package
            while pkg is not None:
                if pkg.plugin is not None:
                    plugin = pkg.plugin
                    if self.typetable.get(plugin.lower()) is None:
                        return "ERROR: Unknown plugin"

                    elif self.typetable[plugin.lower()].get(blocktype) is None:
                        return (
                            "ERROR: Plugin("
                            + plugin
                            + ') does not provide types for "'
                            + blocktype
                            + '"'
                        )

                    typeinfo = self.typetable[plugin.lower()][blocktype].get(
                        ".".join(types).lower()
                    )
                    if typeinfo is None:
                        pkg = pkg.parent
                        continue
                    else:
                        constructor = typeinfo["constructor"]
                        break

                pkg = pkg.parent

        else:
            typeinfo = self.typetable[plugin.lower()][blocktype].get(
                ".".join(types).lower()
            )
            constructor = typeinfo["constructor"]

        return constructor


_TypeTable = {
    "": {"Para": {"import": {"constructor": None}, "access": {"constructor": None}}}
}



# データ構造階層
# ・ Gdoc               文書タイトル、著者などの属性情報を持ち、コンテンツとしてブロックリストを持つ
# ・ BlockList          ソースポジション、オリジナルエレメントタイプを属性にもち、Blockをコンテンツとして持つ
#                       BlockListもBlockであり、つまり階層化する。
# ・ Block
#     ・ Table          テーブル構造。ヘッダ行・ボディ行列をもつ。各行内のセルはBlockList
#     ・ DefinitionList 定義リスト。mdには存在しない。
#     ・ InlineList     ソースポジションを属性にもち、テキスト情報をコンテンツとして持つ
#                       InlineListを子に持ち、階層化するケースがある。
#                       -> InlineList.text = ['Text']   // 配列の各要素が、1行分のテキストを含む。
#
# ・ Inline             LineBlockの構成要素。LineBlockは１つ以上のLineの集合（リスト）であり、
#                       Lineは、ひとつ以上のInlineの集合（リスト）である
#                       InlineはInlineを含むことができ、階層化する

import logging

from gdoc.lib import debug

_LOGGER = logging.getLogger(__name__)
_DEBUG = debug.Debug(_LOGGER)


class Element:
    def __init__(self, pan_elem, elem_type, *, type_def=None):
        self.pan_element = pan_elem
        self.type = elem_type
        self.type_def = type_def if type_def is not None else _PANDOC_TYPES[self.type]
        self.parent = None
        self.children = []
        self.source = _SourcePos(self)

    def _append_child(self, child):
        child.parent = self
        self.children.append(child)

    def next(self):
        next = None

        if self.parent is not None:
            index = self.parent.children.index(self) + 1
            if index < len(self.parent.children):
                next = self.parent.children[index]

        return next

    def prev(self):
        prev = None

        if self.parent is not None:
            index = self.parent.children.index(self) - 1
            if index >= 0:
                prev = self.parent.children[index]

        return prev

    def get_parent(self):

        return self.parent

    def get_children(self):
        return self.children[:]

    def get_first_child(self):
        child = None

        if len(self.children) > 0:
            child = self.children[0]

        return child

    def getFirstChild(self):
        return self.get_first_child()

    def get_type(self):
        return self.type

    def get_prop(self, name):
        TYPEDEF = self.type_def
        property = None

        if ("content" in TYPEDEF) and (TYPEDEF["content"] is not None):
            element = self.pan_element

            if ("key" in TYPEDEF["content"]) and (TYPEDEF["content"]["key"] is not None):
                element = element[TYPEDEF["content"]["key"]]

            if (
                ("struct" in TYPEDEF)
                and (TYPEDEF["struct"] is not None)
                and (name in TYPEDEF["struct"])
            ):
                index = TYPEDEF["struct"][name]

                if isinstance(index, dict):
                    index = index["index"]

                property = element[index]

        return property

    def getProp(self, name):
        return self.get_prop(name)

    def get_attr(self, name):
        attr = None

        attr_obj = self.get_prop("Attr")

        if attr_obj is not None:
            for item in attr_obj[2]:
                if (isinstance(name, str) and (item[0] == name)) or (
                    isinstance(name, tuple) and (item[0] in name)
                ):
                    attr = item[1]
                    break

        return attr

    def hascontent(self):
        TYPEDEF = self.type_def

        hascontent = ("content" in TYPEDEF) and (TYPEDEF["content"] is not None)

        return hascontent

    def get_content(self):
        TYPEDEF = self.type_def
        content = None

        if self.hascontent():
            if ("key" in TYPEDEF["content"]) and (TYPEDEF["content"]["key"] is not None):
                content = self.pan_element[TYPEDEF["content"]["key"]]
            else:
                content = self.pan_element

            if ("main" in TYPEDEF["content"]) and (
                TYPEDEF["content"]["main"] is not None
            ):
                content = content[TYPEDEF["content"]["main"]]

        return content

    def get_content_type(self):
        TYPEDEF = self.type_def
        content_type = None

        if self.hascontent():
            if "type" in TYPEDEF["content"]:
                content_type = TYPEDEF["content"]["type"]

        return content_type

    def walk(self, action, post_action=None, target=None):
        #
        # def action(elem, root):
        #
        element = target or self

        action(element, self)

        if element.children is not None:
            for child in element.children[:]:
                self.walk(action, post_action, child)

        if post_action is not None:
            post_action(element, self)

    @classmethod
    def create_element(cls, elem, elem_type=""):

        elem_type = elem_type or elem["t"]

        gdoc_elem = _PANDOC_TYPES[elem_type]["class"](elem, elem_type)

        return gdoc_elem


class _SourcePos:
    def __init__(self, elem):
        self.position = elem.get_attr(("pos", "data-pos"))


class Inline(Element):
    def __init__(self, pan_elem, elem_type, parent=None):
        super().__init__(pan_elem, elem_type)
        self.text = ""

        _DEBUG.print(elem_type + "（Inline）")
        _DEBUG.indent()

        if not self.hascontent():
            #
            # 'Space', 'SoftBrak' or 'LineBreak'
            #
            self.children = None

            # 暫定。config.jsonから読み込むようにする。
            self.text = {"Space": " ", "SoftBreak": " ", "LineBreak": "\n"}[elem_type]

        elif self.get_content_type() == "Text":
            #
            # In Inline context, 'Text' is a text string
            #
            self.children = None
            self.text = self.get_content()

        else:
            #
            # '[Inline]' or '[Block]'
            #
            panContent = self.get_content()

            for element in panContent:
                self._append_child(Element.create_element(element))

            for element in self.children:
                if hasattr(element, "text") and (element.text is not None):
                    self.text += element.text

        _DEBUG.undent()


class Block(Element):
    def __init__(self, pan_elem, elem_type, parent=None):
        super().__init__(pan_elem, elem_type)


class BlockList(Block):
    def __init__(self, pan_elem, elem_type, parent=None):
        super().__init__(pan_elem, elem_type)

        _DEBUG.print(elem_type + "（BlockList）")
        _DEBUG.indent()

        contents = self.get_content()
        type = "ListItem" if self.get_content_type() == "[[Block]]" else ""

        for item in contents:
            self._append_child(Element.create_element(item, type))

        _DEBUG.undent()

    def getFirstLine(self):
        line = ""
        child = self.get_first_child()

        if child.type == "ListItem":
            child = child.get_first_child()

        if isinstance(child, InlineList):
            line = child.getFirstLine()

        return line


class InlineList(Block):
    def __init__(self, pan_elem, elem_type, parent=None):
        super().__init__(pan_elem, elem_type)
        self.text = []

        _DEBUG.print(elem_type + "（InlineList）")
        _DEBUG.indent()

        if not self.hascontent():
            # HorizontalRule
            self.children = None
            self.text = None

        elif self.get_content_type() == "Text":
            #
            # In LineBlock context, 'Text' is a list of lines.
            #
            self.children = None

            panContent = self.get_content()
            self.text = panContent.split("\n")

        else:
            contents = self.get_content()
            type = "InlineList" if self.get_content_type() == "[[Inline]]" else ""

            for item in contents:
                self._append_child(Element.create_element(item, type))

            lines = ""
            for element in self.children:
                if hasattr(element, "text") and (element.text is not None):
                    lines += element.text

            self.text = lines.split("\n")

        if self.text is not None:
            _DEBUG.puts("-----\n" + "\n".join(self.text) + "\n-----\n")

        _DEBUG.undent()

    def getFirstLine(self):
        line = ""

        if isinstance(self.text, list):
            line = self.text[0]

        return line


class DefinitionList(Block):
    # DefinitionList [([Inline], [[Block]])]
    # - Definition list. Each list item is a pair consisting of a term (a list of inlines) and one or more definitions (each a list of blocks)

    def __init__(self, pan_elem, elem_type):
        super().__init__(pan_elem, elem_type)

        _DEBUG.print(elem_type + "（DefinitionList）")
        _DEBUG.indent()

        if self.get_type() == "DefinitionList":
            # 初入、DL全体。
            # DefinitionList [([Inline], [[Block]])]
            contents = self.get_content()

            for content in contents:
                self._append_child(Element.create_element(content, "DefinitionItem"))

        else:
            # 再入、DLの各項目。
            # DefinitionListItem ([Inline], [[Block]])
            contents = pan_elem

            # [Inline]
            self._append_child(Element.create_element(contents[0], "InlineList"))

            # [[Block]]
            for content in contents[1]:
                # [Block]
                self._append_child(Element.create_element(content, "BlockList"))

        _DEBUG.undent()


class Table(Block):
    # Table Attr Caption [ColSpec] TableHead [TableBody] TableFoot
    # 'c': ['Attr', 'Caption', '[ColSpec]', 'TableHead', '[TableBody]', 'TableFoot']
    def __init__(self, pan_elem, elem_type):
        super().__init__(pan_elem, elem_type)

        _DEBUG.print(elem_type + "（Table）")
        _DEBUG.indent()

        self.numTableColumns = 0  # Num of Columns of Table
        self.numHeaderRows = 0  # Num of Rows of Header
        self.numBodys = 0  # Num of Bodys of Table (NOT Rows. Each Body includes Rows)
        self.numBodyRows = []  # Num of Rows of each Body
        self.numRowHeaderColumns = []  # Num of Columns of RowHeader of each Body
        self.numFooterRows = 0  # Num of Rows of Footer
        self.numTableRows = 0  # Num of Rows of Table
        #   = numHeaderRows + sum(numBodyRows) + numFooterRows
        self.cells = []  # Cell index table

        # Step.1: get num of table columns from len([ColSpec])
        self.numTableColumns = len(self.get_prop("[ColSpec]"))

        # Step.2: Header
        table_head = TableRowList(self.get_prop("TableHead"), "TableHead")
        self._append_child(table_head)
        self.numHeaderRows = table_head.num_rows

        # Step.3: Bodys = [ ( RowHeads, Rows ) ]
        table_bodys = self.get_prop("[TableBody]")
        self.numBodys = len(table_bodys)

        for table_body in table_bodys:
            body = TableBody(table_body, "TableBody")
            self._append_child(body)
            self.numBodyRows.append(body.num_rows)
            self.numRowHeaderColumns.append(body.get_prop("RowHeadColumns"))

        # Step.4: Footer
        table_foot = TableRowList(self.get_prop("TableFoot"), "TableFoot")
        self._append_child(table_foot)
        self.numFooterRows = table_foot.num_rows

        # set up props
        self.numTableRows = (
            self.numHeaderRows + sum(self.numBodyRows) + self.numFooterRows
        )
        self._create_cell_index()

        _DEBUG.undent()

    def _create_cell_index(self):
        # Step.1: Header
        child = self.get_first_child()
        rows = child.children
        for row in rows:
            self.cells.append(row.children[:])

        # Step.2: Bodys = [ ( RowHeads, Rows ) ]
        child = child.next()

        while child.type == "TableBody":
            # Body = ( RowHeads, Rows )
            row_heads = child.get_first_child()
            rows = row_heads.next()
            hashead = child.get_prop("RowHeadColumns") > 0

            # cells of a row = concatenate rowhead and row.
            for index in range(rows.num_rows):
                cells = []
                if hashead:
                    cells.extend(
                        row_heads.children[index].children
                    )  # row_heads -> row -> cells
                cells.extend(rows.children[index].children)  # rows -> row -> cells
                self.cells.append(cells)

            child = child.next()

        # Step.3: Footer
        rows = child.children
        for row in rows:
            self.cells.append(row.children[:])

    def getFirstLine(self):
        line = self.cells[0][0].getFirstLine()
        return line

    def getRow(self, r):
        return self.cells[r - 1]

    def getCell(self, r, c):
        cell = self.cells[r - 1][c - 1]
        return cell


class TableRowList(Element):
    def __init__(self, pan_elem, elem_type, parent=None):
        super().__init__(pan_elem, elem_type)

        _DEBUG.print(elem_type + "（TableRowList）")
        _DEBUG.indent()

        rows = self.get_content()
        self.num_rows = len(rows)

        for row in rows:
            self._append_child(TableRow(row, "Row"))

        _DEBUG.undent()


class TableBody(Element):
    def __init__(self, pan_elem, elem_type, parent=None):
        super().__init__(pan_elem, elem_type)

        _DEBUG.print(elem_type + "（TableBody）")
        _DEBUG.indent()

        rows = self.get_prop("RowHeads")
        self._append_child(TableRowList(rows, "Rows"))

        rows = self.get_prop("Rows")
        self._append_child(TableRowList(rows, "Rows"))

        self.num_rows = len(rows)

        _DEBUG.undent()


class TableRow(Element):
    def __init__(self, pan_elem, elem_type="Row", parent=None):
        super().__init__(pan_elem, elem_type)

        _DEBUG.print(elem_type + "（TableRow）")
        _DEBUG.indent()

        contents = self.get_content()

        for content in contents:
            self._append_child(TableCell(content, "Cell"))

        _DEBUG.undent()

    def getFirstLine(self):
        firstCell = self.getFirstChild()
        line = firstCell.getFirstLine()

        return line

    def getCell(self, c):
        return self.children[c - 1]

    def hasContent(self, start=1, end=-1):
        if end == -1:
            end = len(self.children)

        for c in range(start, end):
            if not self.getCell(c).isEmpty:
                return True

        return False


class TableCell(Element):
    def __init__(self, pan_elem, elem_type, parent=None):
        super().__init__(pan_elem, elem_type)

        _DEBUG.print(elem_type + "（TableCell）")
        _DEBUG.indent()

        self.rowSpan = self.get_prop("RowSpan")
        self.colSpan = self.get_prop("ColSpan")

        blocks = self.get_content()
        for block in blocks:
            self._append_child(Element.create_element(block))

        _DEBUG.undent()

    def getFirstLine(self):
        line = ""
        block = self.get_first_child()

        if isinstance(block, InlineList):
            line = block.getFirstLine()

        return line

    def isEmpty(self):
        empty = len(self.children) == 0
        return empty


class PandocAst(Element):
    def __init__(self, pandoc_ast):
        """
        与えられた PandocAst オブジェクトに対して、以下を行う。
        ・ 各エレメントを走査して、Gdoc apps/pluginsが解釈しやすいデータ構造を提供する
        ・ Source positionのためだけの、付加情報をもたない Div/Span 要素による階層を省く
        ・ Header blockにより文書全体をセクション分割・階層化する
        ・ 各エレメントに、parent へのリンクを付与する
        ・ 文書の装飾情報を省き、apps/plugins が必要な情報をシンプルに提供する
            ・ 連結したテキストデータを提供する
            ・ ブロックエレメントの種類を減らした、シンプルなデータモデルを提供する
            ・ 元文書の装飾やデータタイプ情報にアクセスできる手段を提供する
        """
        super().__init__(pandoc_ast, "Pandoc")

        contents = self.get_content()

        for block in contents:
            self._append_child(Element.create_element(block))

        # Step 2: Remove wrapper 'div'/'span' which contains only one block or is an only child.
        self.walk(self._remove_wrapper)

    def _remove_wrapper(self, elem, root):
        if elem.type in ["Div", "Span"]:

            if len(elem.children) == 0:
                self._remove_elem(elem)

            elif len(elem.children) == 1:
                if elem.source.position is not None:

                    if elem.children[0].source.position is None:
                        elem.children[0].source.position = elem.source.position

                    elif (len(elem.parent.children) == 1) and (
                        elem.parent.source.position is None
                    ):
                        elem.parent.source.position = elem.source.position

                self._remove_elem(elem)

            else:
                if len(elem.parent.children) == 1:
                    if elem.parent.source.position is None:
                        elem.parent.source.position = elem.source.position
                    self._remove_elem(elem)

    def _remove_elem(self, elem):
        index = elem.parent.children.index(elem)

        if len(elem.children) == 0:
            del elem.parent.children[index]

        else:
            if len(elem.children) == 1:
                elem.parent.children[index] = elem.children[0]
                elem.children[0].parent = elem.parent

            elif len(elem.parent.children) == 1:
                elem.parent.children = elem.children[:]
                for e in elem.parent.children:
                    e.parent = elem.parent


# Text.Pandoc.Definition
# Definition of Pandoc data structure for format-neutral representation of documents.
# https://hackage.haskell.org/package/pandoc-types-1.22/docs/Text-Pandoc-Definition.html
#
_PANDOC_TYPES = {
    #
    # Gdoc additional types
    #
    "BlockList": {
        # [Block]   is not BlockList object, just an Array of Blocks.
        #           It means BlockList doesn't have 't' and 'c' elements.
        "class": BlockList,
        "content": {"key": None, "type": "[Block]"},
        "struct": None,
    },
    "InlineList": {
        # [Inline]  is not InlineList object, just an Array of Inlines.
        #           It means InlineList doesn't have 't' and 'c' elements.
        "class": InlineList,
        "content": {"key": None, "type": "[Inline]"},
        "struct": None,
    },
    "ListItem": {
        # [Block]   is not ListItem object, just an Array of Blocks.
        #           It means ListItem doesn't have 't' and 'c' elements.
        "class": BlockList,
        "content": {"key": None, "type": "[Block]"},
        "struct": None,
    },
    "DefinitionItem": {
        # ([Inline], [[Block]]) is not List, Item(=Term+Definitions).
        "class": DefinitionList,
        "content": {
            "key": None
            # 'type':     [ '[Inline]', '[[Block]]' ]
        },
        "struct": None,
    },
    #
    # Pandoc
    #
    "Pandoc": {
        # Pandoc Meta [Block]
        "class": BlockList,
        "content": {"key": None, "main": "blocks", "type": "[Block]"},
        "struct": {"Version": "pandoc-api-version", "Meta": "meta", "Blocks": "blocks"},
    },
    #
    # Blocks
    #
    "Plain": {
        # Plain [Inline]
        # - Plain text, not a paragraph
        "class": InlineList,
        "content": {"key": "c", "type": "[Inline]"},
        "struct": None,
    },
    "Para": {
        # Para [Inline]
        # - Paragraph
        "class": InlineList,
        "content": {"key": "c", "type": "[Inline]"},
        "struct": None,
    },
    "LineBlock": {
        # LineBlock [[Inline]]
        # - Multiple non-breaking lines
        "class": InlineList,
        "content": {"key": "c", "type": "[[Inline]]"},
        "struct": None,
    },
    "CodeBlock": {
        # CodeBlock Attr Text
        # - Code block (literal) with attributes
        "class": InlineList,
        "content": {"key": "c", "type": "Text"},
        "struct": {"Attr": 0, "Text": 1},
    },
    "RawBlock": {
        # RawBlock Format Text
        # - Raw block
        "class": InlineList,
        "content": {"key": "c", "main": 1, "type": "Text"},
        "struct": {"Format": 0, "Text": 1},
    },
    "BlockQuote": {
        # BlockQuote [Block]
        # - Block quote (list of blocks)
        "class": BlockList,
        "content": {"key": "c", "type": "[Block]"},
        "struct": None,
    },
    "OrderedList": {
        # OrderedList ListAttributes [[Block]]
        # - Ordered list (attributes and a list of items, each a list of blocks)
        "class": BlockList,
        "content": {"key": "c", "main": 1, "type": "[[Block]]"},
        "struct": {"ListAttributes": 0, "ListItems": 1},
    },
    "BulletList": {
        # BulletList [[Block]]
        # - Bullet list (list of items, each a list of blocks)
        "class": BlockList,
        "content": {"key": "c", "type": "[[Block]]"},
        "struct": None,
    },
    "DefinitionList": {
        # DefinitionList [([Inline], [[Block]])]
        # - Definition list. Each list item is a pair consisting of a term (a list of inlines) and one or more definitions (each a list of blocks)
        "class": DefinitionList,
        "content": {"key": "c", "type": "[([Inline], [[Block]])]"},
        "struct": None,
    },
    "Header": {
        # Header Int Attr [Inline]
        # - Header - level (integer) and text (inlines)
        "class": InlineList,
        "content": {"key": "c", "main": 2, "type": "[Inline]"},
        "struct": {"Level": 0, "Attr": 1, "[Inline]": 2},
    },
    "HorizontalRule": {
        # HorizontalRule
        # - Horizontal rule
        "class": InlineList
    },
    "Table": {
        # Table Attr Caption [ColSpec] TableHead [TableBody] TableFoot
        # - Table, with attributes, caption, optional short caption, column alignments and widths (required), table head, table bodies, and table foot
        "class": Table,
        "content": {"key": "c", "main": None},
        "struct": {
            "Attr": 0,
            "Caption": 1,
            "[ColSpec]": 2,
            "TableHead": 3,
            "[TableBody]": 4,  # TableBody*s*
            "TableFoot": 5,
        },
    },
    "Div": {
        # Div Attr [Block]
        # - Generic block container with attributes
        "class": BlockList,
        "content": {"key": "c", "main": 1, "type": "[Block]"},
        "struct": {"Attr": 0, "[Block]": 1},
    },
    "Null": {
        # Null
        # - Nothing
    },
    #
    # Inlines
    #
    "Str": {
        # Str Text
        # Text (string)
        "class": Inline,
        "content": {"key": "c", "type": "Text"},
        "struct": None,
    },
    "Emph": {
        # Emph [Inline]
        # Emphasized text (list of inlines)
        "class": Inline,
        "content": {"key": "c", "type": "[Inline]"},
        "struct": None,
    },
    "Underline": {
        # Underline [Inline]
        # Underlined text (list of inlines)
        "class": Inline,
        "content": {"key": "c", "type": "[Inline]"},
        "struct": None,
    },
    "Strong": {
        # Strong [Inline]
        # Strongly emphasized text (list of inlines)
        "class": Inline,
        "content": {"key": "c", "type": "[Inline]"},
        "struct": None,
    },
    "Strikeout": {
        # Strikeout [Inline]
        # Strikeout text (list of inlines)
        "class": Inline,
        "content": {"key": "c", "type": "[Inline]"},
        "struct": None,
    },
    "Superscript": {
        # Superscript [Inline]
        # Superscripted text (list of inlines)
        "class": Inline,
        "content": {"key": "c", "type": "[Inline]"},
        "struct": None,
    },
    "Subscript": {
        # Subscript [Inline]
        # Subscripted text (list of inlines)
        "class": Inline,
        "content": {"key": "c", "type": "[Inline]"},
        "struct": None,
    },
    "SmallCaps": {
        # SmallCaps [Inline]
        # Small caps text (list of inlines)
        "class": Inline,
        "content": {"key": "c", "type": "[Inline]"},
        "struct": None,
    },
    "Quoted": {
        # Quoted QuoteType [Inline]
        # Quoted text (list of inlines)
        "class": Inline,
        "content": {"key": "c", "main": 1, "type": "[Inline]"},
        "struct": {"QuotedType": 0, "Inlines": 1},
    },
    "Cite": {
        # Cite [Citation] [Inline]
        # Citation (list of inlines)
        "class": Inline,
        "content": {"key": "c", "main": 1, "type": "[Inline]"},
        "struct": {"Citation": 0, "Inlines": 1},
    },
    "Code": {
        # Code Attr Text
        # Inline code (literal)
        "class": Inline,
        "content": {"key": "c", "main": 1, "type": "Text"},
        "struct": {"Attr": 0, "Text": 1},
    },
    "Space": {
        # Space
        # Inter-word space
        "class": Inline
    },
    "SoftBreak": {
        # SoftBreak
        # Soft line break
        "class": Inline
    },
    "LineBreak": {
        # LineBreak
        # Hard line break
        "class": Inline
    },
    "Math": {
        # Math MathType Text
        # TeX math (literal)
        "class": Inline,
        "content": {"key": "c", "main": 1, "type": "Text"},
        "struct": {"MathType": 0, "Text": 1},
    },
    "RawInline": {
        # RawInline Format Text
        # Raw inline
        "class": Inline,
        "content": {"key": "c", "main": 1, "type": "Text"},
        "struct": {"Format": 0, "Text": 1},
    },
    "Link": {
        # Link Attr [Inline] Target
        # Hyperlink: alt text (list of inlines), target
        "class": Inline,
        "content": {"key": "c", "main": 1, "type": "[Inline]"},
        "struct": {"Attr": 0, "Inlines": 1, "Target": 2},
    },
    "Image": {
        # Image Attr [Inline] Target
        # Image: alt text (list of inlines), target
        "class": Inline,
        "content": {"key": "c", "main": 1, "type": "[Inline]"},
        "struct": {"Attr": 0, "Inlines": 1, "Taget": 2},
    },
    "Note": {
        # Note [Block]
        # Footnote or endnote
        "class": Inline,
        "content": {"key": "c", "type": "[Block]"},
        "struct": None,
    },
    "Span": {
        # Span Attr [Inline]
        # Generic inline container with attributes
        "class": Inline,
        "content": {"key": "c", "main": 1, "type": "[Inline]"},
        "struct": {"Attr": 0, "Inlines": 1},
    },
    #
    # OtherTypes
    #
    "TableHead": {
        # TableHead Attr [Row]
        # The head of a table.
        "class": TableRowList,
        "content": {"key": None, "main": 1, "type": "[Row]"},
        "struct": {"Attr": 0, "Rows": 1},
    },
    "TableBody": {
        # TableBody Attr RowHeadColumns [Row] [Row]
        # A body of a table, with an intermediate head, intermediate body,
        # and the specified number of row header columns in the intermediate body.
        "class": TableBody,
        "content": {"key": None},
        "struct": {
            "Attr": 0,
            "RowHeadColumns": 1,
            "RowHeads": {"index": 2, "type": "[Row]"},
            "Rows": {"index": 3, "type": "[Row]"},
        },
    },
    "TableFoot": {
        # TableFoot Attr [Row]
        # The foot of a table.
        "class": TableRowList,
        "content": {"key": None, "main": 1, "type": "[Row]"},
        "struct": {"Attr": 0, "Rows": 1},
    },
    "Rows": {
        # Row Attr [Cell]
        # A table row.
        "class": TableRowList,
        "content": {"key": None, "type": "[Row]"},
        "struct": None,
    },
    "Row": {
        # Row Attr [Cell]
        # A table row.
        "class": TableRow,
        "content": {"key": None, "main": 1, "type": "[Cell]"},
        "struct": {
            # 'Attr':     0,    Commented out because of issue about handling Rows in Table.
            "Cells": 1
        },
    },
    "Cell": {
        # Cell Attr Alignment RowSpan ColSpan [Block]
        # A table cell.
        "class": TableCell,
        "content": {"key": None, "main": 4, "type": "[Block]"},
        "struct": {
            "Attr": 0,
            "Alignment": 1,
            "RowSpan": 2,
            "ColSpan": 3,
            "[Block]": {"index": 4, "type": "[Block]"},
        },
    },
}



r"""
Block class
"""

# from .types import create_element
from .block import Block


class BlockList(Block):
    """
    BlockList is a Block containing Blocks or BlockLists as a list.
    """

    def __init__(self, pan_elem, elem_type, type_def, create_element):
        """Constructor
        @param pan_elem(Dict)
            PandocAST Element
        @param elem_type(Str)
            Element type
        @param type_def(Dict)
            Element structur data
        @param create_element(func(pan_elem, elem_type : Element))
            General constructor of pandoc Element types for creating children.
        """
        super().__init__(pan_elem, elem_type, type_def)

        contents = self.get_content()
        type = self.get_content_type()

        for item in contents:
            self._add_child(create_element(item, type))



r"""
Element class
"""

from typing import List, Union


class Element:
    """
    Base class of PandocAST Element handler.
    """

    def __init__(self, pan_elem: dict, elem_type: str, type_def: dict):
        """Constructor
        @param pan_elem(Dict)
            PandocAST Element
        @param elem_type(Str)
            Element type
        @param type_def(Dict)
            Element structur data
        """
        self.pan_element: dict = pan_elem
        self.type: str = elem_type
        self.type_def: dict = type_def
        self.parent: Union[Element, None] = None
        self.children: List[Element] = []

    def _add_child(self, child) -> "Element":
        """add an element as a child.
        @param child(Element)
            Element to add as a child.
        @return Element :
            self, for chaining.
        """
        child.parent = self
        self.children.append(child)
        return self

    def next(self) -> Union["Element", None]:
        """returns an element ordered at next to self.
        @return Element :
            If the next element does not exist, returns None.
        """
        next = None

        if self.parent is not None:
            index = self.parent.children.index(self) + 1
            if index < len(self.parent.children):
                next = self.parent.children[index]

        return next

    def prev(self) -> Union["Element", None]:
        """returns an element ordered at previous to self.
        @return Element :
            If the previous element does not exist, returns None.
        """
        prev = None

        if self.parent is not None:
            index = self.parent.children.index(self) - 1
            if index >= 0:
                prev = self.parent.children[index]

        return prev

    def get_parent(self) -> Union["Element", None]:
        """returns parent element.
        @return Element : The parent element.
        """
        return self.parent

    def get_children(self) -> Union[List["Element"], None]:
        """returns new copied list of child elements.
        @return [Element] :
            new list copied from children.
        """
        children = None

        if self.children is not None:
            children = self.children[:]

        return children

    def get_first_child(self) -> Union["Element", None]:
        """returns the first child elements.
        @return Element : The first child.
        """
        child = None

        if len(self.children) > 0:
            child = self.children[0]

        return child

    def get_type(self) -> str:
        """returns element type.
        @return Str : Type string.
        """
        return self.type

    def get_prop(self, key):
        """returns a property of the element specified by key string.
        @param key(Str)
            Key string of the property.
        @return Str :
            Value string of the property.
        """
        TYPEDEF = self.type_def
        property = None

        if ("content" in TYPEDEF) and (TYPEDEF["content"] is not None):
            element = self.pan_element

            if ("key" in TYPEDEF["content"]) and (TYPEDEF["content"]["key"] is not None):
                element = element[TYPEDEF["content"]["key"]]

            if (
                ("struct" in TYPEDEF)
                and (TYPEDEF["struct"] is not None)
                and (key in TYPEDEF["struct"])
            ):
                index = TYPEDEF["struct"][key]

                if isinstance(index, dict):
                    index = index["index"]

                property = element[index]

        return property

    def get_attr(self, name):
        """returns a attribute of the element specified by key string.
        @param key(Str or tuple)
            Key string of the attribute.
            Multiple keys can be specified in tuple format.
            In this case, it returns the value of the first attribute found among the keys in the tuple.
        @return Str :
            Value string of the attribute.
        """
        attr = None

        attr_obj = self.get_prop("Attr")

        if attr_obj is not None:
            for item in attr_obj[2]:
                if (isinstance(name, str) and (item[0] == name)) or (
                    isinstance(name, tuple) and (item[0] in name)
                ):
                    attr = item[1]
                    break

        return attr

    def hascontent(self) -> bool:
        """returns True if self has content(s) or False if self is typed but has no content.
        @return Bool :
        """
        TYPEDEF = self.type_def

        hascontent = ("content" in TYPEDEF) and (TYPEDEF["content"] is not None)

        return hascontent

    def get_content(self):
        """returns main content data in the element.
        @return Main content data of the element.
        """
        TYPEDEF = self.type_def
        content = None

        if self.hascontent():
            if ("key" in TYPEDEF["content"]) and (TYPEDEF["content"]["key"] is not None):
                content = self.pan_element[TYPEDEF["content"]["key"]]
            else:
                content = self.pan_element

            if ("main" in TYPEDEF["content"]) and (
                TYPEDEF["content"]["main"] is not None
            ):
                content = content[TYPEDEF["content"]["main"]]

        return content

    def get_content_type(self) -> Union[str, None]:
        """returns type of main content in the element.
        @return String : The type of main content in the element.
        """
        TYPEDEF = self.type_def
        content_type = None

        if self.hascontent():
            if "type" in TYPEDEF["content"]:
                content_type = TYPEDEF["content"]["type"]

        return content_type

    def walk(self, action, post_action=None, opt=None) -> "Element":
        """Walk through all elements of the tree and call out given functions.
        @param action(function) : def action(element, opt)
        @param post_action(function) : def post_action(element, opt)
        @return Element :
            self, for chaining.
        """
        action(self, opt)

        if self.children is not None:
            for child in self.children[:]:
                child.walk(action, post_action, opt)

        if post_action is not None:
            post_action(self, opt)

        return self

    #
    # '_item()' related methods
    #

    def get_parent_item(self, ignore=["Div", "Span"]) -> Union["Element", None]:
        """returns parent item.
        @return Element : The parent item.
        """
        parent = self.parent

        while parent is not None:
            if parent.type not in ignore:
                break

            parent = parent.parent

        return parent

    def get_child_items(self, ignore=["Div", "Span"]) -> List["Element"]:
        """returns new copied list of child items.
        @return [Element] :
        """
        child_items: List[Element] = []

        if self.children is None:
            return None

        for child in self.children:
            if child.type in ignore:
                items = child.get_child_items(ignore)
                if items is not None:
                    child_items += items
            else:
                child_items.append(child)

        return child_items

    def next_item(self, ignore=["Div", "Span"]) -> Union["Element", None]:
        """returns an element ordered at next to self.
        @return Element :
            If the next element does not exist, returns None.
        """
        next: Union[Element, None] = None

        parent = self.get_parent_item(ignore)
        if parent is not None:
            siblings = parent.get_child_items(ignore)
            index = siblings.index(self) + 1
            if index < len(siblings):
                next = siblings[index]

        return next

    def prev_item(self, ignore=["Div", "Span"]) -> Union["Element", None]:
        """returns an element ordered at previous to self.
        @return Element :
            If the previous element does not exist, returns None.
        """
        prev: Union[Element, None] = None

        parent = self.get_parent_item(ignore)
        if parent is not None:
            siblings = parent.get_child_items(ignore)
            index = siblings.index(self) - 1
            if index >= 0:
                prev = siblings[index]

        return prev

    def get_first_item(self, ignore=["Div", "Span"]) -> Union["Element", None]:
        """returns the first child item.
        @return Element : The first item.
        """
        first: Union[Element, None] = None

        items = self.get_child_items(ignore)
        if items is not None and len(items) > 0:
            first = items[0]

        return first

    def walk_items(
        self, action, post_action=None, opt=None, ignore=["Div", "Span"]
    ) -> "Element":
        """Walk through all items of the tree and call out given functions.
        @param action(function) : def action(element, opt)
        @param post_action(function) : def post_action(element, opt)
        @return Element :
            self, for chaining.
        """
        if self.type not in ignore:
            action(self, opt)

        if self.children is not None:
            for child in self.children[:]:
                child.walk_items(action, post_action, opt, ignore)

        if post_action is not None and self.type not in ignore:
            post_action(self, opt)

        return self



r"""
Pandoc class
"""
import json
import subprocess


class Pandoc:
    """
    Execute external pandoc command as a subprocess to parse a source md file to generate PandocAST json object.
    """

    def __init__(self):
        """Constructor"""
        self._version = None
        self._version_str = None

    def _run(self, commandlines, stdin=None):
        """run multiple subcommands
        run multiple subcommands while connecting it with pipes and return the output.
        @param comandlines([Str])
            Commandline strings including options to run as subprocesses.
        @param stdin
            stdin port passed to the first subcommand.
        @return output : Str
            stdout from the last subcommand.
        """
        output = None
        args = commandlines[0].split()

        with subprocess.Popen(
            args, stdin=stdin, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        ) as ps:

            if len(commandlines) > 1:
                output = self._run(commandlines[1:], stdin=ps.stdout)
                ps.wait()

            else:
                ps.wait()
                output = b"".join(ps.stdout.readlines())

            if ps.returncode != 0:
                # subprocess.check_output():
                # https://docs.python.org/3/library/subprocess.html#subprocess.check_output
                # > If the return code was non-zero it raises a CalledProcessError.
                raise subprocess.CalledProcessError(
                    ps.returncode,
                    args,
                    b"".join(ps.stdout.readlines()),
                    b"".join(ps.stderr.readlines()),
                )

        return output

    def get_json(self, filepath: str, formatType=None, html=None):
        """
        returns pandoc ast json object.
        @param filepath : str
            File path to the source document to read.
        @param fromType : str / None
            Input filetype to be passed to pandoc with opt '-f'.
        @param html : bool / None
            Flag to specify whether or not to interpret html tags.
        @return output : { PandocAst Object }
            Pandoc AST json object generated by pandoc from source document.
        """
        _SOURCEPOS_ = ["gfm", "commonmark", "commonmark_x"]

        filename = filepath.split("/")[-1]
        file_ext = filename.split(".")
        file_ext = file_ext[-1].lower() if len(file_ext) > 1 else ""

        if formatType is None and html is None:
            if file_ext == "md":
                formatType = "gfm+sourcepos"
                html = True
            else:
                formatType = None
                html = False
        else:
            if html is None:
                html = False

            if formatType in _SOURCEPOS_:
                formatType += "+sourcepos"

        commands = []
        cmd = "pandoc " + filepath

        if formatType is not None:
            cmd += " -f " + formatType

        if not html:
            cmd += " -t json"
            commands.append(cmd)
        else:
            cmd += " -t html"
            commands.append(cmd)
            commands.append("pandoc -f html -t json")

        json_object = json.loads(self._run(commands))

        return json_object

    def get_version(self):
        """
        returns versions of pandoc and pandoc-types.
        @return versions : {
                'pandoc': [int or str],
                'pandoc-types': [int or str],
            }
            version information obtained from 'pandoc --version'
        """
        if self._version is not None:
            return self._version

        version_str = self._run(["pandoc --version"]).decode()
        lines = version_str.split("\n")

        version = {}
        for line in lines:
            words = line.replace(",", " ").split()

            for i, word in enumerate(words):
                if ((word == "pandoc") or (word == "pandoc-types")) and (
                    i < len(words) - 1
                ):
                    vers = words[i + 1].split(".")
                    if not vers[0].isdecimal():
                        break

                    version[word] = []
                    for ver in vers:
                        version[word].append(int(ver) if ver.isdecimal() else ver)

        if ("pandoc" in version) and ("pandoc-types" in version):
            self._version = version
            self._version_str = version_str

        return self._version



from collections.abc import Iterable
from typing import Union, cast

from gdoc.lib.gdoc import String, Text, TextString

from ..fsm import State

_TOKEN_CHARS = (" ", ",", "=", "@", "[", "]", "(", ")", "{", "}", '"', "\\", "\n")


def tokenize_textstring(textstr: TextString) -> TextString:
    tokenizer: State = cast(State, TextTokenizer().start().on_entry())

    for text in textstr:
        if isinstance(text, String):
            text = cast(Iterable, text)
            for c in text:
                tokenizer.on_event(c)

        else:
            tokenizer.on_event(text)

    return tokenizer.on_exit()


class TextTokenizer(State):
    """
    TextTokenizer
    """

    def on_entry(self):
        next = self

        self.tokens = TextString()
        self.word = String()

        return next

    def on_event(self, event: Union[Text, String]):
        next = self

        if TextTokenizer.is_word(event):
            self.word += event

        else:
            self._flush_word_buff()
            self.tokens.append(event)

        return next

    def on_exit(self):
        self._flush_word_buff()
        return self.tokens

    def _flush_word_buff(self):
        if len(self.word) > 0:
            # Flush buffer
            self.tokens.append(self.word[:])
            self.word = String()

    @classmethod
    def is_word(cls, token):
        return isinstance(token, String) and (token[0] not in _TOKEN_CHARS)



"""
document.py: Document class
"""

import copy

from ..pandocastobject.pandocast import PandocAst
from .config import DEFAULTS
from .section import Section


class Document(Section):
    """ """

    def __init__(self, pandoc: PandocAst = None, opts={}):
        self.__opts = copy.deepcopy(DEFAULTS)
        self.__opts.update(opts)

        ignore = self.__opts.get("pandocast", {}).get("types", {}).get("ignore")
        kwargs = {"ignore": ignore} if ignore is not None else {}
        blocks = pandoc.get_child_items(**kwargs)
        super().__init__(blocks, opts=self.__opts)



"""
blocktagparser.py: parse_BlockTag function
"""
from typing import Any, Optional, TypeAlias, Union, cast

from gdoc.lib.gdoc import String, Text, TextString
from gdoc.lib.gdoccompiler.gdexception import GdocSyntaxError
from gdoc.util import Err, Ok, Result
from gdoc.util.errorreport import ErrorReport

from ..fsm import State, StateMachine
from ..textblock.texttokenizer import TextTokenizer

ObjectTagInfo: TypeAlias = tuple[
    tuple[String | None, String | None, String | None],
    #    [category,      type,          isref        ]
    list[TextString],
    list[tuple[TextString, TextString]],
]


def parse_ObjectTagInfo(
    tagstr: TextString, opts: dict, erpt: ErrorReport
) -> Result[ObjectTagInfo, ErrorReport]:
    """
    parse_ObjectTag
    """
    class_info: tuple[String | None, String | None, String | None] = (None, None, None)
    #           tuple[category, type, is_reference]
    class_args: list[TextString] = []
    class_kwargs: list[tuple[TextString, TextString]] = []
    tokens: TextString = tagstr[:]

    if len(tokens) > 0:
        # parse Class Info
        if TextTokenizer.is_word(tokens[0]):
            # get class info
            class_info, e = parse_ClassInfo(tokens[0], opts, erpt)
            if e:
                erpt.submit(e)
                return Err(erpt)

            tokens = tokens[1:]

        elif not (isinstance(tokens[0], String) and (tokens[0] == " ")):
            # Class info should be String
            erpt.submit(GdocSyntaxError())
            return Err(erpt)

        args, e = parse_Arguments(tokens, opts, erpt)
        if e:
            erpt.submit(e)
            return Err(e)

        class_args, class_kwargs = args

    return Ok((class_info, class_args, class_kwargs))


def parse_ClassInfo(
    token: String, opts: dict, erpt: ErrorReport
) -> Result[tuple[String | None, String | None, String | None], ErrorReport]:
    #       tuple[category, type, is_reference]
    class_cat: String | None = None
    class_type: String | None = None
    class_isref: String | None = None

    if (i := token.find(":")) >= 0:
        class_cat = token[:i]
        class_type = token[i + 1 :]

        if cast(Union[str, String], class_type).find(":") >= 0:
            raise GdocSyntaxError()

    else:
        class_type = token[:]

    if cast(Union[str, String], class_type).endswith("&"):
        class_isref = cast(String, class_type)[-1]
        class_type = cast(String, class_type)[:-1]

    return Ok((class_cat, class_type, class_isref))


#
# Argument Parser
#
def parse_Arguments(
    tokens: TextString, opts: dict, erpt: ErrorReport
) -> Result[tuple, ErrorReport]:
    """
    element ::=
    [ word | quoted TextString | ( arguments ) | "=" | " " | "," | OtherTextTypeTokens ]*
    """
    elements, e = detect_parentheses(tokens, opts, erpt)
    if e:
        erpt.submit(e)
        return Err(erpt)

    parser = ArgumentParser().start()
    parser.on_entry()

    for e in elements:
        parser.on_event(e)

    parser.on_event(None)  # EOL

    return Ok(parser.on_exit())


class ArgumentParser(StateMachine):
    """
    returns argument

    argument = {
        "args": [ value ],
        "kwargs": [
            [ key, value ]
        ]
    }

    element ::= [ word | "quoted" | (bracketed) | "=" | " " | "," | NotStringTokens ]*

    Key: TextString ::= word    # should be isidentifier()
    Value: TextString ::= [ word | "quoted" | (bracketed) | NotStringTokens ]*
    """

    def __init__(self, name: str = None) -> None:
        super().__init__(name)
        self.add_state(_Idle("Idle"), None)
        self.add_state(_Key("Key"), "AfterKey")
        self.add_state(_AfterKey("AfterKey"), None)
        self.add_state(_Value("Value"), "Idle")

    def start(self, param=None):
        self.argstring: TextString = TextString()
        self.args: list[Text] = []
        self.kwargs: list[list[Text]] = []
        return super().start((self.argstring, self.args, self.kwargs))

    def on_entry(self, event=None):
        return super().on_entry(event)

    def on_event(self, token):
        return super().on_event(token)

    def on_exit(self):
        super().on_exit()
        return self.args, self.kwargs


class _Idle(State):
    """ """

    def start(self, param: Any = None) -> "State":
        self.comma: Union[String, bool, None] = None
        return super().start(param)

    def on_entry(self, element=None):
        next = self

        if self.comma is not None:
            self.comma = False

        if element == ",":
            self.comma = element

        elif element is not None:
            next = self.on_event(element)

        return next

    def on_event(self, element):
        next = self

        if element in (" ", None):
            pass

        elif element == ",":
            if self.comma is False:
                self.comma = element
            else:
                raise GdocSyntaxError()

        else:
            next = ("Key", element)

        return next


class _Key(State):
    """ """

    def start(self, param):
        self.argstring: TextString
        self.args: list[Text]
        self.kwargs: list[list[Text]]

        self.argstring, self.args, self.kwargs = param

    def on_entry(self, element):
        self.argstring.clear()
        self.argstring.append(element)
        return self

    def on_event(self, element):
        next = self

        if element is None:
            if len(self.argstring) > 0:
                self.args.append(self.argstring[:])

        elif element in (" ", ",", "="):
            next = (None, element)

        else:
            self.argstring.append(element)

        return next


class _AfterKey(State):
    """ """

    def start(self, param):
        self.argstring: TextString
        self.args: list[Text]
        self.kwargs: list[list[Text]]

        self.argstring, self.args, self.kwargs = param

    def on_entry(self, element=None):
        next = self
        if element:
            next = self.on_event(element)

        return next

    def on_event(self, element):
        next = self
        kwargs: bool = len(self.kwargs) > 0

        if element == " ":
            pass

        elif element == "=":
            next = "Value"

        elif element is None:
            if len(self.argstring) > 0:
                self.args.append(self.argstring[:])

        else:
            if kwargs:
                raise GdocSyntaxError()

            self.args.append(self.argstring[:])
            next = ("Idle", element)

        return next


class _Value(State):
    """ """

    def start(self, param):
        self.argstring: TextString
        self.kwargs: list[list[Text]]

        self.argstring, _, self.kwargs = param

        self.value: TextString = TextString()

    def on_entry(self, element=None):
        self.value.clear()

        if element:
            self.value.append(element)

        return self

    def on_event(self, element):
        next = self

        if element is None:
            pass
            # todo: if value is empty, rase error.

        elif element in (" ", ",", "="):
            if len(self.value) > 0:
                next = (None, element)

        else:
            self.value.append(element)

        return next

    def on_exit(self):
        self.kwargs.append((self.argstring[:], self.value[:]))
        return


def detect_parentheses(
    tokens: TextString, opts: dict, erpt: ErrorReport
) -> Result[TextString, ErrorReport]:
    elements: TextString = TextString()

    detector = ParenthesesDetector().start(elements)

    try:
        detector.on_entry()

        for token in tokens:
            detector.on_event(token)

        detector.on_exit()

    except GdocSyntaxError as e:
        erpt.submit(e)
        return Err(erpt)

    return Ok(elements)


class ParenthesesDetector(StateMachine):
    """ """

    def __init__(self, name: str = None) -> None:
        super().__init__(name)
        self.add_state(_Main("Main"), None)
        self.add_state(_Parentheses("Parentheses"), "Main")

    def start(self, param):
        self.elements = param
        return super().start(self.elements)

    def on_entry(self, event=None):
        return super().on_entry(event)

    def on_exit(self):
        super().on_exit()


class _Main(State):
    """ """

    def __init__(self, name: str = None) -> None:
        super().__init__(name)
        self.elements: Optional[TextString] = None

    def start(self, elements):
        self.elements = elements

    def on_event(self, token):
        next = self

        if isinstance(token, String) and (token == "("):
            next = ("Parentheses", token)

        elif isinstance(token, String) and (token == ")"):
            raise GdocSyntaxError()
            # It may be better to call State "Parentheses" to raise

        else:
            self.elements.append(token)

        return next


class _Parentheses(State):
    """ """

    def __init__(self, name: str = None) -> None:
        super().__init__(name)
        self.elements: Optional[TextString] = None
        self._tokens: Optional[TextString] = None

    def start(self, elements):
        self.elements = elements

    def on_entry(self, token):
        self._tokens = TextString()
        self._tokens.append(token)  # always it's '('.
        self._parencount = 0
        return self

    def on_event(self, token):
        next: Union[State, None, tuple[str, tuple[int, Text]]] = self

        self._tokens.append(token)

        if isinstance(token, String) and (token == "("):
            self._parencount += 1

        elif isinstance(token, String) and (token == ")"):
            if self._parencount > 0:
                self._parencount -= 1
            else:
                next = None

        return next

    def on_exit(self):
        self.elements.append(self._tokens)
        self._tokens = None



"""
textblockparser.py: parse_TextBlock function
"""
from gdoc.lib.gdoc import String, Text, TextBlock, TextString
from gdoc.lib.gdoc.blocktag import BlockTag, BlockTagInfo
from gdoc.lib.gdoccompiler.gdexception import GdocSyntaxError
from gdoc.lib.gobj.types import BaseObject
from gdoc.util import Err, Ok, Result

from ....util.errorreport import ErrorReport
from .lineparser import parse_Line


def parse_TextBlock(
    textblock: TextBlock, gobj: BaseObject, opts: dict, erpt: ErrorReport
) -> Result[BaseObject | None, ErrorReport]:
    """
    parse TextBlock and creates Gobj.

    @param textblock (TextBlock) : _description_
    @param gobj (BaseObject) : _description_
    @param opts (dict) : _description_
    @param erpt (ErrorReport) : _description_

    @return Result[BaseObject, ErrorReport] : if created, returns the new TextObject.
                                        othrewise, None.
    """
    srpt = erpt.new_subreport()

    parsed_lines: list[TextString] = []
    line: TextString
    parsed_line: TextString
    for line in textblock:
        parsed_line, e = parse_Line(line, opts, srpt)
        if e and srpt.submit(e):
            return Err(srpt)

        if parsed_line:
            parsed_lines.append(parsed_line)

    preceding_lines: list[TextString] = []
    following_lines: list[TextString] = []
    preceding_text: TextString | None = None
    following_text: TextString | None = None
    block_tag: BlockTag | None = None
    for parsed_line in parsed_lines:
        #
        # TODO: Add a check for double tagging
        #
        if block_tag is None:
            i: int
            text: Text
            for i, text in enumerate(parsed_line):
                if isinstance(text, BlockTag):
                    block_tag = text
                    break

            if block_tag is None:
                preceding_lines.append(parsed_line)
            else:
                preceding_text = parsed_line[:i]
                following_text = parsed_line[i + 1 :]
        else:
            following_lines.append(parsed_line)

    child: BaseObject | None = None
    if block_tag is not None:
        class_info: tuple[String | None, String | None, String | None]
        class_args: list[TextString]
        class_kwargs: list[tuple[TextString, TextString]]
        class_info, class_args, class_kwargs = block_tag.get_class_arguments()

        block_tag.tag_info = BlockTagInfo(
            textblock, preceding_lines, following_lines, preceding_text, following_text
        )

        tag_opts = _get_blocktag_opts(
            preceding_lines, following_lines, preceding_text, following_text
        )

        try:
            child = gobj.create_object(
                class_info,
                class_args,
                class_kwargs,
                tag_opts,
                block_tag,
            )

        except GdocSyntaxError as e:
            srpt.submit(e)
            return Err(srpt)

    if srpt.haserror():
        return Err(srpt, child)

    return Ok(child)


def _get_blocktag_opts(
    preceding_lines: list[TextString],
    following_lines: list[TextString],
    preceding_text: TextString | None,
    following_text: TextString | None,
) -> dict[str, TextString | list[TextString] | None]:
    """
    1. The following text of btag will be used as the name. \
        Ignore comment-outs by `[]` is not support yet.
    2. If the preceding text has one or more `-` words at the end,
        then Preceding lines + Preceding text (excluding `-`) will be the property "text".
    3. If 2 is False, Following lines will be set to the property "text".
    4. If the text contains inline tags, add the properties specified by the tags
        without deleting from content of text property.
    """
    tag_opts: dict[str, TextString | list[TextString] | None] = {}

    #
    # Set `name`
    #
    tag_opts["name"] = None
    if type(following_text) is TextString:
        name = following_text.strip()
        if len(name) > 0:
            tag_opts["name"] = name

    #
    # Set postposition tag
    #
    tag_opts["text"] = None
    if type(preceding_text) is TextString:
        pretext: TextString = preceding_text.rstrip()
        hyphen: str | None = None
        text: TextString | None = None

        for i in range(len(pretext)):
            if hyphen is None:
                if pretext[-1 - i] == "-":
                    hyphen = "-"
                else:
                    break

            elif hyphen == "-":
                if pretext[-1 - i] == "-":
                    continue
                elif pretext[-1 - i].isspace():
                    hyphen == " "
                else:
                    break

            elif hyphen == " ":
                if pretext[-1 - i].isspace():
                    continue
                else:
                    break

        if hyphen == " ":
            text = preceding_lines[:]
            text.append(pretext[:-i])
        else:
            text = following_lines[:]

        if text:
            tag_opts["text"] = text

    return tag_opts



r"""
PandocAst class
"""

from .blocklist import BlockList


class Pandoc(BlockList):
    """
    Root element representing whole pandocAst object.
    """

    def __init__(self, pan_elem, elem_type, type_def, create_element):
        """Constructor
        @param pan_elem(Dict)
            PandocAST Element
        @param elem_type(Str)
            Element type
        @param type_def(Dict)
            Element structur data
        @param create_element(func(pan_elem, elem_type : Element))
            General constructor of pandoc Element types for creating children.
        """
        super().__init__(pan_elem, elem_type, type_def, create_element)



r"""
Plugin class
"""


class Category:
    """ """

    def __init__(self, category_info):
        """ """
        self.name = category_info["name"]
        self.version = category_info["version"]
        self.module = category_info["module"]
        self.types = category_info["types"]
        self.aliases = category_info["aliases"]
        self.defaults = category_info["defaults"]

        for t in set(self.types.values()):
            if hasattr(t, "set_category"):
                t.set_category(self)

    def get_type(self, target_type: str, parent_type: str, opts: dict = {}):
        """
        get object class by name
        """
        constructor = None
        type_name = None

        # Get default type if target_type is not provided
        if target_type in ("", None):
            if ("defaults" in opts) and (parent_type in opts["defaults"]):
                target_type = opts["defaults"][parent_type]

            elif parent_type in self.defaults:
                target_type = self.defaults[parent_type]

        if target_type not in ("", None):
            type_name = target_type.upper()

            if ("aliases" in opts) and (type_name in opts["aliases"]):
                type_name = opts["aliases"][type_name]

            if type_name in self.aliases:
                type_name = self.aliases[type_name]

            if type_name in self.types:
                constructor = self.types[type_name]

        return type_name, constructor



r"""
Gdoc Compiler class
"""
import os

from gdoc.lib.gdoc import Document as GdocDocument
from gdoc.lib.gdocparser.documentparser import parse_Document
from gdoc.lib.gobj.types.document import Document as GobjDocument
from gdoc.lib.pandocastobject.pandoc import Pandoc
from gdoc.lib.pandocastobject.pandocast import PandocAst
from gdoc.util import Err, ErrorReport, Ok, Result


class GdocCompiler:
    """ """

    def __init__(self) -> None:
        pass

    def compile(
        self, filepath: str, opts: dict = {}
    ) -> Result[GobjDocument, ErrorReport | Exception]:
        """
        1. fileの存在確認
        2. ./_gdoc_/filename.past.json の存在確認
        3. pandocの実行
        4. pandocAstの生成
        5. metadataの取得 → Documentにセット
        6. パーサーの生成、visitorの取得、start(Document)。
        7. Pandoc.accept(parser)して、エレメントをイベントとしてパーサーに投げる。
        """
        erpt = ErrorReport(cont=opts.get("check-only"))

        if not os.path.isfile(filepath):
            erpt.submit(
                # should be Exception
                f"{filepath} is not found."
            )
            return Err(erpt)

        pandoc_json = Pandoc().get_json(filepath, "gfm+sourcepos", False)
        pandoc_ast = PandocAst(pandoc_json)
        gdoc = GdocDocument(pandoc_ast)

        gobj = GobjDocument(None, filepath)

        gobj, e = parse_Document(gdoc, gobj, opts, erpt)
        if e:
            erpt.submit(e)
            return Err(erpt, gobj)

        return Ok(gobj)



"""
fsm.py: Finite State Machine
"""
from typing import Any, Dict, List, Optional, Tuple, Union, cast


class State:
    def __init__(self, name: str = None) -> None:
        self.name = name or self.__class__.__name__

    def start(self, param: Any = None) -> "State":
        return self  # `self` for chaining

    def on_entry(
        self, event: Any = None
    ) -> Union[Union["State", str, None], Tuple[Union["State", str, None], Any]]:
        # continue      --> self
        # re-entry      --> (self, event)
        # transition(forward)
        #               --> state or (state, event)
        # done(default transition)
        #               --> None or (None, result)
        return self

    def on_event(
        self, event: Any
    ) -> Union[Union["State", str, None], Tuple[Union["State", str, None], Any]]:
        # continue      --> self
        # re-entry      --> (self, event)
        # transition    --> state or (state, event)
        # done(default transition)
        #               --> None or (None, result)
        return self

    def on_exit(self) -> Any:
        return

    def stop(self) -> Any:
        return


class StateMachine(State):
    def __init__(self, name: str = None) -> None:
        super().__init__(name or self.__class__.__name__)

        self.__state_list: List[State] = []
        self.__next_state: Dict["State", Union["State", str, None]] = {}
        self.__current_state: Optional["State"] = None

    def add_state(
        self, state: State, next: Union["State", str, None] = None
    ) -> "StateMachine":
        if isinstance(state, State):
            self.__state_list.append(state)
            self.__next_state[state] = next
        else:
            raise TypeError(
                'state to add should be State or StateMachine(not "'
                + type(state).__name__
                + '")'
            )

        return self  # `self` for chaining

    def start(self, param=None) -> "StateMachine":
        self.__current_state = self.__state_list[0]

        for state in self.__state_list:
            state.start(param)

        return self  # `self` for chaining

    def on_entry(
        self, event=None
    ) -> Union[Union["State", str, None], Tuple[Union["State", str, None], Any]]:
        # continue      --> self
        # re-entry      --> (self, event)
        # transition(forward)
        #               --> state or (state, event)
        # done(default transition)
        #               --> None or (None, result)
        next: Union[
            Union["State", str, None], Tuple[Union["State", str, None], Any]
        ] = None

        if self.__current_state is None:
            # Not yet started.
            raise RuntimeError("StateMachine " + self.name + " is stopped / not started.")

        next = self.__current_state.on_entry(event)
        return self.__move_to(next)

    def on_event(
        self, event
    ) -> Union[Union["State", str, None], Tuple[Union["State", str, None], Any]]:
        # continue      --> self
        # re-entry      --> (self, event)
        # transition    --> state or (state, event)
        # done(default transition)
        #               --> None or (None, result)
        if self.__current_state is None:
            # Not yet started.
            raise RuntimeError("StateMachine " + self.name + " is stopped / not started.")

        next = self.__current_state.on_event(event)
        return self.__move_to(next)

    def on_exit(self) -> Any:
        if self.__current_state is not None:
            self.__current_state.on_exit()

    def stop(self) -> Any:
        #
        # Should be guaranteed that on_exit() will be called here.
        #
        for state in self.__state_list:
            state.stop()

        self.__current_state = None

    def __move_to(
        self, next
    ) -> Union[Union["State", str, None], Tuple[Union["State", str, None], Any]]:
        # continue      --> self
        # re-entry      --> (self, event)
        # transition    --> state or (state, event)
        # done(default transition)
        #               --> None or (None, result)
        if next is self.__current_state:
            # continue
            next = self  # `self` is __current_state for parent state

        else:
            cast(State, self.__current_state).on_exit()

            # Unpack
            if type(next) is tuple:
                _next, _event = next
            else:
                _next = next
                _event = None

            # Get default next state
            if _next is None:
                _next = self.__next_state[cast(State, self.__current_state)]

            # done, and no next state
            if _next is None:
                self.__current_state = None
                # next = None or (None, event)

            # move to next
            else:
                self.__current_state = self.__get_state(_next)

                if self.__current_state is not None:
                    next = self.__current_state.on_entry(_event)
                    next = self.__move_to(next)

                # else:
                # if self.current_state is None(Not found),
                # return the unfound state to upper layer.
                # next = next or (next, event)

        return next

    def __get_state(self, next) -> Optional[State]:
        state = None

        if isinstance(next, State):
            if next in self.__state_list:
                state = next

        elif type(next) is str:
            for s in self.__state_list:
                if s.name == next:
                    state = s
                    break

        elif next is not None:
            raise RuntimeError("Returned next state is not State")

        return state



r"""
Package class
"""

from gdoc.lib.gdoccompiler.gdexception import *

from .baseobject import BaseObject


class Package(BaseObject):
    """ """

    def __init__(self, typename, id, *, scope="+", name=None, tags=[], **kwargs):
        """ """

    def add_document(self, symbol, cat_name=None, type_name=None, scope="+", **kwargs):
        """ """

    def add_package(self):
        """ """



import logging
import subprocess

_LOGGER = logging.getLogger(__name__)


class Pandoc:
    def __init__(self) -> None:
        pass

    def run(self, file, fromType=None):
        commands = []

        if fromType is None:
            if (file.split(".")[-1]) == "md":
                cmd = "pandoc -f gfm+sourcepos -t html".split()
                cmd.append(file)
                commands.append(cmd)

                cmd = "pandoc -f html -t json".split()
                commands.append(cmd)

            else:
                cmd = "pandoc -t json".split()
                cmd.append(file)
                commands.append(cmd)

        else:
            cmd = "pandoc -t json -f".split()
            cmd.append(fromType)
            cmd.append(file)
            commands.append(cmd)

        output = None
        procs = []
        with subprocess.Popen(commands[0], stdout=subprocess.PIPE) as ps:
            procs.append(ps)

            if len(commands) > 1:
                with subprocess.Popen(
                    commands[1], stdin=procs[0].stdout, stdout=subprocess.PIPE
                ) as ps:
                    procs.append(ps)
                    output = procs[-1].communicate()[0]
                    procs[0].communicate()
                    if (procs[0].returncode != 0) or (procs[1].returncode != 0):
                        # should raise
                        output = None

            else:
                output = procs[-1].communicate()[0]
                if procs[0].returncode != 0:
                    # should raise
                    output = None

        return output



"""
gdoc: Gdoc data structure
"""
from .document import Document
from .section import Section
from .string import String
from .text import TEXT, Text
from .textblock import TextBlock
from .textstring import TextString

__all__ = ["Text", "TEXT", "String", "TextString", "TextBlock", "Section", "Document"]



# SysML: System Description Language plugin
#
# 1. Requirement diagram 相当の情報を取り扱う
#

import logging

from ...lib import debug
from ...lib.gdsymbol import Symbol
from ...lib.symboltable import Scope
from ...lib.types.table.hierarchicalDict import HierarchicalDict

_LOGGER = logging.getLogger(__name__)
_DEBUG = debug.Debug(_LOGGER)

_version = "0.3.0"


class Requirement(HierarchicalDict):
    def __init__(self, table, tag=None, parent=None) -> None:
        _DEBUG.print("class Requirement(HierarchicalDict) {")
        _DEBUG.indent()

        super().__init__(table, tag)

        self.setProp(None, "plugin", "sysml")
        self.setProp(None, "type", "requirement")
        self.setProp(None, "version", _version)

        self.itemTable = self._createItems()

        _DEBUG.undent()
        _DEBUG.print("}")

    def getItems(self, symboltable):
        items = []

        for item in self.itemTable:
            items.append(item.getItem(symboltable))

        return items

    def _createItems(self, element=None, parentId=""):
        items = []

        elements = self.getChildren(element)
        for key in elements:
            ids = key.split(".")
            if len(ids) > 1:
                if not parentId.endswith(".".join(ids[:-1])):  # 現在、tagを無視している
                    # should raise
                    return None
            id = "" if parentId == "" else parentId + "."
            id += ids[-1]
            # 事前準備
            items.append(RequirementItem(self, id, elements[key], parentId))
            items += self._createItems(elements[key], id)  # 現在、tagを無視している

        _DEBUG.print("createItems: len(items) = " + str(len(items)))
        return items

    def link(self):

        for item in self.itemTable:
            for linktype in item.trace:
                for linkto in item.trace[linktype]:
                    linkitem = item.symboltable.resolve(linkto)
                    if linkitem is not None:
                        if not linktype in item.link["to"]:
                            item.link["to"][linktype] = []
                        item.link["to"][linktype].append(linkitem)

                        if not linktype in linkitem.link["from"]:
                            linkitem.link["from"][linktype] = []
                        linkitem.link["from"][linktype].append(item)

                    else:
                        if not linktype in item.link["to"]:
                            item.link["to"][linktype] = []
                        item.link["to"][linktype].append(linkto)


class RequirementItem:
    def __init__(self, reqt, id, element, parentId="") -> None:
        self.symboltable = None

        self.id = Symbol.getId(id)
        self.tags = Symbol.getTags(id)
        self.name = reqt.getProp(None, "Name", element)
        self.stereotype = reqt.getProp(None, "Name", element)
        self.text = reqt.getProp(None, "Description", element)
        self.trace = Trace(reqt.getProp(None, "Trace", element)).content
        self.assosiation = {}
        self.link = {"to": {}, "from": {}}

        if parentId != "":
            self.addTrace("deriveReqt", parentId)

    # addItem(self, symbol, name, objectClass, item, scope=Scope.PUBLIC):
    def getItem(self, symboltable=None):
        if symboltable is not None:
            self.symboltable = symboltable

        item = {}
        item["symbol"] = self.id
        item["name"] = self.name
        item["objectClass"] = RequirementItem
        item["item"] = self
        item["scope"] = Scope.PUBLIC

        return item

    def addTrace(self, type, id):
        if self.trace.get(type) is None:
            self.trace[type] = []
        self.trace[type].append(id)


class Trace:
    def __init__(self, table) -> None:
        self.content = {}
        if isinstance(table, list):
            if isinstance(table[0], list):
                data = []
                for row in table:
                    data = data + row
            else:
                data = table
        elif isinstance(table, str):
            data = [table]
        else:
            # should raise
            return None

        for cell in data:
            self._parseTrace(self.content, cell)

    def _parseTrace(self, content, cell):
        result = {}
        key = "trace"
        _VALID_TRACE_TYPES = {
            "trace": "trace",
            "copy": "copy",
            "refine": "refine",
            "derive": "deriveReqt",
            "derivereqt": "deriveReqt",
        }

        cell = cell.replace(",", " ")
        for word in cell.split():
            if word.startswith("@"):
                word = word[1:].lower()
                if _VALID_TRACE_TYPES.get(word) is None:
                    # should raise
                    _LOGGER.warning("trace type '%s' is invalid.", word)
                    key = "_INVALID_"
                else:
                    key = _VALID_TRACE_TYPES.get(word)
            else:
                if Symbol.isValid(word):
                    if not key in content:
                        content[key] = []

                    content[key].append(word)
                else:
                    # should raise
                    _LOGGER.warning("trace id '%s' is invalid.", word)



"""
string.py: String class
"""

from typing import Optional

from gdoc.lib.pandocastobject.pandocast import PandocAst, PandocElement
from gdoc.lib.pandocastobject.pandocstr import PandocStr

from .text import Text


class String(PandocStr, Text):
    """
    ImmutableSequence of Character strings of PandocAst inline elements.
    """

    def __init__(
        self,
        items: Optional[PandocStr | str | list[PandocElement]] = None,
        start: int = 0,
        stop: int = None,
    ):
        if isinstance(items, PandocStr):
            super().__init__()
            self += items[start:stop]

        elif type(items) is str:
            super().__init__(
                [PandocAst.create_element({"t": "Str", "c": items})], start, stop
            )

        else:
            super().__init__(items, start, stop)

    def get_str(self) -> str:
        return str(self)

    def get_text(self) -> str:
        return self.get_str()

    def __getitem__(self, *args, **kwargs):
        return String(super().__getitem__(*args, **kwargs))

    def strip(self, *args, **kwargs):
        return String(super().strip(*args, **kwargs))

    def lstrip(self, __chars: Optional[str] = None) -> "String":
        return String(super().lstrip(__chars))

    def rstrip(self, __chars: Optional[str] = None) -> "String":
        return String(super().rstrip(__chars))



from enum import Enum, auto

from .gdsymbol import Symbol


class Scope(Enum):
    PUBLIC = auto()
    PRIVATE = auto()


class SymbolTable:
    def __init__(self, owner, parent) -> None:
        self.owner = owner  # Package
        self.parent = parent  # SymbolTable
        self.table = {}  # ObjectItem or SymbolTable

        self.fullId = owner.id
        if (parent is not None) and (parent.fullId != ""):
            self.fullId = parent.fullId + "." + self.fullId

    def addItem(self, symbol, name, objectClass, item, scope=Scope.PUBLIC):

        id = Symbol.getId(symbol)
        tags = Symbol.getTags(symbol)

        if id in self.table:
            return "ERROR: duplicate id"

        self.table[id] = {
            "tags": tags,
            "name": name,
            "scope": scope,
            "objectClass": objectClass,
            "item": item,
        }

    def resolve(self, symbol):
        fp = Symbol.getFilePath(symbol)
        if fp is not None:
            return None  # not yet implemented.

        id = Symbol.getId(symbol)
        ids = id.split(".")
        item = None

        item = self._resolve(id, ids)

        if (item is None) and (self.parent is not None):
            item = self.parent._resolve(id, ids)

        return item

    def _resolve(self, id, ids):

        item = None

        if id in self.table:
            item = self.table[id]["item"]

        elif (len(ids) > 1) and (ids[0] in self.table):
            child = self.table[ids[0]]["item"]
            if isinstance(child, SymbolTable):
                ids = ids[1:]  # Copy List not to affect to original argument.
                item = child._resolve(".".join(ids), ids)

        return item

    def search(self, id):
        return self._search(id, id.split("."))

    def _search(self, id, ids):
        foundItems = []

        foundItem = self._resolve(id, ids)
        if foundItem is not None:
            foundItems.append(foundItem)

        for key in self.table:
            item = self.table[key]

            if isinstance(item["item"], SymbolTable):
                foundItems = foundItems + item["item"]._search(id, ids)

        return foundItems

    def dump(self):
        data = self.table.copy()
        for key in data:
            item = self.table[key]

            if isinstance(item["item"], SymbolTable):
                item["items"] = item["item"].dump()

            del item["item"]
            del item["objectClass"]
            item["scope"] = "Public" if (item["scope"] == Scope.PUBLIC) else "Private"

        return data



"""
text.py: Text class
"""
from abc import ABC, abstractmethod
from typing import TypeVar

TEXT = TypeVar("TEXT", bound="Text")


class Text(ABC):
    """
    Abstract base class for gdoc inline Elements.
    """

    @abstractmethod
    def get_str(self) -> str:
        """
        Return a human readable string.

        A subclass inheriting from `Text` returns its own content
        in a readable string format.

        - Ex. Code("some code").get_str() returns "`some code`"
          (added surrounding char "`").

        @return human readable string of this object.
        """

    @abstractmethod
    def get_text(self) -> str:
        """
        Return the text data contained in this element.

        - Ex. Code("some code").get_text() returns "some code"
          (original text data without surrounding char "`").

        @return the text data contained in this element.
        """



r"""!

**primitive types**

"""
import sys

from .baseobject import BaseObject, ClassInfo
from .importobject import ImportObject

CATEGORY_INFO = {
    "name": "",
    "version": "",
    "module": sys.modules[__name__],
    "types": {"OBJECT": BaseObject, "IMPORT": ImportObject, "ACCESS": ImportObject},
    "aliases": {
        # "OBJ": "OBJECT"
    },
    "defaults": {"": "OBJECT", "OBJECT": "OBJECT", "DOCUMENT": "OBJECT"},
}



"""
textstring.py: TextString class
"""

from typing import Optional, SupportsIndex, Union, overload

from gdoc.lib.pandocastobject.pandocast import PandocElement

from .code import Code
from .string import String
from .text import Text


class TextString(list[Text], Text):
    """
    MutableSequence of gdoc inline elements.

    - @trace(realize): `_BlockId_`
    """

    def __init__(
        self,
        items: Union[list[PandocElement], list[Text], str, "TextString"] = [],  # type: ignore
        opts={}
        # if this > items: list[Element] | list[Text] | str | "TextString" = []
        # returns > E   TypeError: unsupported operand type(s) for |
        # @3.10.7 >              : 'types.UnionType' and 'str'
    ):
        super().__init__()  # init as an empty list

        _plain: list = opts.get("pandocast", {}).get("types", {}).get("plaintext", [])
        _other_known_types = [
            # "Str",
            # "Space",
            # "SoftBreak",
            # "LineBreak",
            # "Code",
            "Math",
            "Image",
            "Quoted",
            "Cite",
            "RawInline",
            "Note",
        ]

        inlines: "TextString" | list[PandocElement] | list[Text] | list[String]

        if type(items) is str:
            inlines = [String(items)]

        elif type(items) is __class__:  # type: ignore
            inlines = items

        elif type(items) is list:
            inlines = items

        else:
            raise TypeError()

        for item in inlines:
            if isinstance(item, Text):
                self.append(item)

            elif isinstance(item, PandocElement):
                etype = item.get_type()

                if etype in _plain:
                    self.append(String([item]))

                elif etype == "LineBreak":
                    self.append(String([item]))

                elif etype == "Code":
                    self.append(Code(item))

                elif etype in _other_known_types:
                    # Not yet supported element types.
                    pass

                else:
                    raise TypeError()
            else:
                raise TypeError()

    # @Override(list)
    def append(self, text: Text) -> None:
        """
        Append `Text` to the end of the list.

        @param text (Text) : text to be added
        @exception TypeError : text is not subclass of `Text`.
        """
        if not isinstance(text, Text):
            raise TypeError()

        super().append(text)

    # @Override(list)
    def __add__(self, *args, **kwargs) -> "TextString":
        return TextString(super().__add__(*args, **kwargs))

    # @Override(list)
    # def __iadd__

    #
    # @Override(list)
    #
    @overload
    def __getitem__(self, __i: SupportsIndex) -> Text:
        ...

    @overload
    def __getitem__(self, __s: slice) -> "TextString":
        ...

    def __getitem__(self, index: SupportsIndex | slice):
        """
        x.__getitem__(y) <==> x[y]

        @param index (SupportsIndex | slice) : _description_
        @return Text : if type(index) is SupportIndex
        @return TextString : if type(index) is slice
        """
        result = super().__getitem__(index)
        if type(result) is list:
            result = TextString(result)

        return result

    # @Override(Text(ABC))
    def get_str(self) -> str:
        result = ""
        for text in self:
            result += text.get_str()

        return result

    # @Override(Text(ABC))
    def get_text(self) -> str:
        result = ""
        for text in self:
            result += text.get_text()

        return result

    #
    # Original `str`-like methods
    #
    def startswith(self, __prefix: str | tuple[str, ...]) -> bool:
        """
        S.startswith(prefix[, start[, end]]) -> bool

        Return True if S(leading `String`s) starts with the specified prefix,
        False otherwise.

        @param __prefix (str | tuple[str, ...]) : Prefix(es)
        @return bool : True if S starts with the specified prefix, False otherwise.
        """
        return self.__get_leading_str().startswith(__prefix)

    def __get_leading_str(self) -> str:
        result: str = ""
        for text in self:
            if type(text) is String:
                result += str(text)
            elif type(text) is TextString:
                string: str = text.__get_leading_str()
                result += string
                if len(string) < len(text.get_text()):
                    break
            else:
                break

        return result

    def endswith(self, __suffix: str | tuple[str, ...]) -> bool:
        """
        S.endswith(suffix) -> bool

        Return True if S(last `String`s) ends with the specified suffix, False otherwise.

        @param __suffix (str | tuple[str, ...]) : Suffix(es)
        @return bool : True if S ends with the specified suffix, False otherwise.
        """
        return self.__get_last_str().endswith(__suffix)

    def __get_last_str(self) -> str:
        result: str = ""
        for text in reversed(self):
            if type(text) is String:
                result = str(text) + result
            elif type(text) is TextString:
                string: str = text.__get_leading_str()
                result = result + string
                if len(string) < len(text.get_text()):
                    break
            else:
                break
        return result

    def strip(self, __chars: Optional[str] = None) -> "TextString":
        """
        Return a copy of the TextString with leading and trailing whitespace removed.

        If chars is given and not None, remove characters in chars instead.

        @param __chars (Optional[str]) : Characters to remove. Defaults to None.

        @return TextString : Copy of the TextString with leading and trailing
                             whitespace removed.
        """.strip()
        return self.rstrip(__chars).lstrip(__chars)

    def lstrip(self, __chars: Optional[str] = None) -> "TextString":
        """
        Return a copy of the TextString with leading whitespace removed.

        If chars is given and not None, remove characters in chars instead.

        @param __chars (Optional[str]) : Characters to remove. Defaults to None.

        @return TextString : Copy of the TextString with leading whitespace removed.
        """
        result: TextString = self[:]

        while (len(result) > 0) and (type(result[0]) in (String, TextString)):
            s: String = result[0]
            result[0] = s.lstrip(__chars)
            if len(result[0]) > 0:
                break
            else:
                del result[0]

        return result

    def rstrip(self, __chars: Optional[str] = None) -> "TextString":
        """
        Return a copy of the TextString with trailing whitespace removed.

        If chars is given and not None, remove characters in chars instead.

        @param __chars (Optional[str]) : Characters to remove. Defaults to None.

        @return TextString : Copy of the TextString with trailing whitespace removed.
        """
        result: TextString = self[:]

        while (len(result) > 0) and (type(result[-1]) in (String, TextString)):
            s: String | TextString = result[-1]
            result[-1] = s.rstrip(__chars)
            if len(result[-1]) > 0:
                break
            else:
                del result[-1]

        return result



r"""
Inline class
"""

from .element import Element


class Inline(Element):
    """
    Inline class of PandocAST Element handler.
    """

    def __init__(self, pan_elem, elem_type, type_def, create_element):
        """Constructor
        @param pan_elem(Dict)
            PandocAST Element
        @param elem_type(Str)
            Element type
        @param type_def(Dict)
            Element structur data
        @param create_element(func(pan_elem, elem_type : Element))
            General constructor of pandoc Element types for creating children.
        """
        super().__init__(pan_elem, elem_type, type_def)
        self.text = ""

        if not self.hascontent():
            #
            # 'Space', 'SoftBrak' or 'LineBreak'
            #
            self.children = None
            self.text = type_def["alt"]

        elif self.get_content_type() == "Text":
            #
            # In Inline context, 'Text' is a text string
            #
            self.children = None
            self.text = self.get_content()

        else:
            #
            # '[Inline]' or '[Block]'
            #
            panContent = self.get_content()
            panContentType = self.get_content_type()

            for element in panContent:
                self._add_child(create_element(element, panContentType))

            for element in self.children:
                if hasattr(element, "text") and (element.text is not None):
                    self.text += element.text



r"""
Gdoc Exception Classes
"""


class GdocIdError(KeyError):
    pass


class GdocKeyError(KeyError):
    pass


class GdocImportError(ImportError):
    pass


class GdocModuleNotFoundError(ModuleNotFoundError):
    pass


class GdocNameError(NameError):
    pass


class GdocRuntimeError(RuntimeError):
    pass


class GdocSyntaxError(SyntaxError):
    pass


class GdocTypeError(TypeError):
    pass
